import pandas as pd
from clinicedc_constants import FEMALE, MALE

from ..constants import MEAN_95CI, N_ONLY, N_WITH_COL_PROP, N_WITH_ROW_PROP
from ..row import RowDefinition, RowDefinitions
from ..table import Table


class FbgTable(Table):
    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(colname="fbg", main_df=main_df, title="FBG (mmol/L) categories")

    @property
    def row_definitions(self) -> RowDefinitions:
        df_tmp = self.main_df.copy()
        row_defs = RowDefinitions(reverse_rows=False)
        row0 = RowDefinition(
            title=self.title,
            label=self.default_sublabel,
            condition=(df_tmp["gender"].notna()),
            columns={FEMALE: (N_ONLY, 2), MALE: (N_ONLY, 2), "All": (N_ONLY, 2)},
            drop=False,
        )
        row_defs.add(row0)

        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="Not fasted",
                condition=(self.main_df["fasting_fbg_hrs"] < 8.0),
                columns=columns,
                drop=True,
            )
        )

        columns = {
            FEMALE: (MEAN_95CI, 2),
            MALE: (MEAN_95CI, 2),
            "All": (MEAN_95CI, 2),
        }
        row_mean = RowDefinition(
            colname=self.colname,
            label="Mean (95% CI)",
            condition=(self.main_df[self.colname].notna()),
            columns=columns,
            drop=False,
        )
        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }

        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="<6.1",
                condition=(self.main_df[self.colname] < 6.1),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="6.1-6.9",
                condition=(self.main_df[self.colname] >= 6.1)
                & (self.main_df[self.colname] < 7.0),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="7.0 and above",
                condition=(self.main_df[self.colname] >= 7.0),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="not measured",
                condition=(self.main_df[self.colname].isna()),
                columns=columns,
                drop=True,
            )
        )

        row_defs.add(row_mean)
        return row_defs



from clinicedc_constants import NOT_APPLICABLE
from django.db import models

from edc_constants.choices import YES_NO_NA


class FollowupReviewModelMixin(models.Model):
    care_delivery = models.CharField(
        verbose_name=(
            "Was care for this `condition` delivered in an integrated care clinic today?"
        ),
        max_length=25,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
        help_text="Select `not applicable` if site was not selected for integrated care.",
    )

    care_delivery_other = models.TextField(
        verbose_name="If NO, please explain", null=True, blank=True
    )

    class Meta:
        abstract = True



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.utils import timezone

from ..exceptions import AllocationError

if TYPE_CHECKING:
    from ..models import StockRequest


def allocate_stock(
    stock_request: StockRequest,
    allocation_data: dict[str, str],
    allocated_by: str,
    user_created: str = None,
    created: datetime = None,
) -> tuple[list[int], list[str]]:
    """Link stock instances to subjects.

    Model `Allocation` is a fkey on Stock and links the stock obj to a
    subject.

    allocation_data: dict of {stock code:subject_identifier} coming from
    the view.

    for any stock instance, the container must be a container used for
    subjects, e.g. bottle 128. That is container__may_request_as=True.

    See post() in AllocateToSubjectView.
    """
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    allocation_model_cls = django_apps.get_model("edc_pharmacy.allocation")
    registered_subject_model_cls = django_apps.get_model("edc_registration.registeredsubject")
    allocated, skipped = [], []
    stock_objs = []
    for code, subject_identifier in allocation_data.items():
        # get rs
        rs_obj = registered_subject_model_cls.objects.get(
            subject_identifier=subject_identifier
        )
        # get the stock request item from this request for this subject
        stock_request_item = stock_request.stockrequestitem_set.filter(
            registered_subject=rs_obj,
            allocation__isnull=True,
        ).first()
        if not stock_request_item:
            skipped.append(f"{subject_identifier}: N/A")
            continue
        # try to create the allocation instance and update the stock instance
        # for this stock code
        try:
            stock_obj = stock_model_cls.objects.get(
                code=code,
                confirmation__isnull=False,
                container__may_request_as=True,
                allocation__isnull=True,
            )
        except ObjectDoesNotExist:
            skipped.append(f"{subject_identifier}: {code}")
        else:
            with transaction.atomic():
                allocation = allocation_model_cls.objects.create(
                    stock_request_item=stock_request_item,
                    code=stock_obj.code,
                    registered_subject=rs_obj,
                    allocation_datetime=timezone.now(),
                    allocated_by=allocated_by,
                    user_created=user_created,
                    created=created,
                )
                # check stock assigment matches subject`s assignment
                if (
                    stock_model_cls.objects.get(code=code).product.assignment
                    != allocation.get_assignment()
                ):
                    allocation.delete()
                    raise AllocationError(
                        "Assignment mismatch. Stock must match subject assignment. "
                        f"Allocation abandoned. See {subject_identifier} and {stock_obj}."
                    )

                stock_obj.allocation = allocation
                stock_obj.allocated = True
                stock_obj.user_modified = user_created
                stock_obj.modified = created
                stock_objs.append(stock_obj)
    if stock_objs:
        for obj in stock_objs:
            obj.save()
            allocated.append(obj.code)
    return allocated, skipped


__all__ = ["allocate_stock"]



from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch.dispatcher import receiver

from edc_lab.utils import get_panel_model
from edc_model.utils import is_inline_model

from ..rule import RuleRunner
from .query_rule import QueryRule

DATA_MANAGER_ENABLED = getattr(settings, "DATA_MANAGER_ENABLED", True)


@receiver(post_save, sender=QueryRule, weak=False, dispatch_uid="update_query_text")
def update_query_text(sender, instance, raw, **kwargs):
    if not raw:
        if not instance.query_text:
            instance.query_text = instance.rendered_query_text or "query not described"
            instance.save_base(update_fields=["query_text"])


@receiver(post_save, weak=False, dispatch_uid="update_query_on_crf")
def update_query_on_crf(sender, instance, raw, **kwargs):
    if not raw:
        if DATA_MANAGER_ENABLED:
            try:
                related_visit_model_attr = instance.related_visit_model_attr()
            except AttributeError:
                pass
            else:
                related_visit = getattr(instance, related_visit_model_attr, None)
                appointment = getattr(related_visit, "appointment", None)
                if appointment and not is_inline_model(instance):
                    visit_schedule_name = appointment.visit_schedule_name
                    opts = dict(
                        visit_schedule__visit_schedule_name=visit_schedule_name,
                        visit_schedule__schedule_name=appointment.schedule_name,
                        visit_schedule__visit_code=appointment.visit_code,
                    )
                    try:
                        instance.panel
                    except AttributeError:
                        opts.update(data_dictionaries__model=sender._meta.label_lower)
                    else:
                        opts.update(
                            requisition_panel=get_panel_model().objects.get(
                                name=instance.panel.name
                            )
                        )
                    for query_rule_obj in (
                        QueryRule.objects.filter(**opts)
                        .order_by("data_dictionaries__model")
                        .distinct()
                    ):
                        runner = RuleRunner(query_rule_obj)
                        runner.run_one(
                            subject_identifier=appointment.subject_identifier,
                            visit_schedule_name=appointment.visit_schedule_name,
                            schedule_name=appointment.schedule_name,
                            visit_code=appointment.visit_code,
                            timepoint=appointment.timepoint,
                        )



from __future__ import annotations

import contextlib
import sys
from pprint import pprint
from typing import TYPE_CHECKING, Any

from django.apps import apps as django_apps
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import QuerySet

from .constants import ACCOUNT_MANAGER_ROLE

if TYPE_CHECKING:
    from django.contrib.auth.models import Group, User

    from .models import Role


def get_user(username: str) -> User | None:
    try:
        user = get_user_model().objects.get(username=username)
    except ObjectDoesNotExist:
        user = None
    return user


def compare_codenames_for_group(group_name: str, expected: list[str]) -> None:
    group = django_apps.get_model("auth.group").objects.get(name=group_name)
    codenames = [p.codename for p in group.permissions.all()]
    new_expected = []
    for c in expected:
        with contextlib.suppress(IndexError):
            c = c.split(".")[1]
        new_expected.append(c)

    compared = [c for c in new_expected if c not in codenames]
    if compared:
        sys.stdout.write(f"{group.name},  missing from codenames\n")
        pprint(compared)  # noqa: T203
    compared = [c for c in codenames if c not in new_expected]
    if compared:
        sys.stdout.write(f"{group.name},  extra codenames\n")
        pprint(compared)  # noqa: T203


def remove_default_model_permissions_from_edc_permissions(auth_updater: Any, app_label: str):
    for group in auth_updater.group_model_cls.objects.all():
        auth_updater.remove_permissions_by_codenames(
            group=group,
            codenames=[
                f"{app_label}.add_edcpermissions",
                f"{app_label}.change_edcpermissions",
                f"{app_label}.delete_edcpermissions",
                f"{app_label}.view_edcpermissions",
            ],
        )


def make_view_only_group_permissions(prefix: str, group: Group, model: str | None = None):
    """Remove all but view permissions for model.

    Accepts a prefix as well, e.g. `historical`.

    Default removes all except `view`.
    """

    opts = dict(codename__contains=f"_{prefix}")
    if model:
        opts.update(model=model)
    for permission in group.permissions.filter(**opts).exclude(
        codename__startswith=f"view_{prefix}"
    ):
        group.permissions.remove(permission)


def get_codenames_for_role(role_name: str) -> list[str]:
    codenames = []
    role_cls = django_apps.get_model("edc_auth.role")
    try:
        roles = role_cls.objects.get(name=role_name).groups.all()
    except ObjectDoesNotExist:
        pass
    else:
        for group in roles:
            codenames.extend(
                [
                    f"{permission.content_type.app_label}.{permission.codename}"
                    for permission in group.permissions.all()
                ]
            )
    return codenames


def user_has_change_perms(**kwargs) -> list[str]:
    codenames = get_codenames_for_user(**kwargs)
    return [c for c in codenames if "add_" in c or "change_" in c or "delete_" in c]


def get_codenames_for_user(
    user: User = None, roles: QuerySet[Role] = None, include_groups: bool | None = None
) -> list[str]:
    codenames: list[str] = []
    groups: list[Group] = []
    account_manager_groups: list[Group] = []
    try:
        role: Role = django_apps.get_model("edc_auth.role").objects.get(
            name=ACCOUNT_MANAGER_ROLE
        )
    except ObjectDoesNotExist:
        pass
    else:
        account_manager_groups: list[Group] = [grp for grp in role.groups.all()]
    roles = roles or user.userprofile.roles

    for role in roles.all():
        groups.extend([grp for grp in role.groups.all() if grp not in account_manager_groups])
    if include_groups:
        groups.extend([grp for grp in user.groups.all() if grp not in account_manager_groups])
    groups = list(set(groups))
    for group in groups:
        codenames.extend(
            [
                f"{permission.content_type.app_label}.{permission.codename}"
                for permission in group.permissions.all()
            ]
        )
    codenames.extend(
        [
            f"{permission.content_type.app_label}.{permission.codename}"
            for permission in user.user_permissions.all()
        ]
    )
    return codenames



# Generated by Django 6.0 on 2026-01-13 00:57

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0093_alter_confirmationatlocation_options_and_more"),
    ]

    operations = []



from django.db import models
from django.utils import timezone

from edc_sites.model_mixins import SiteModelMixin


class ResultItemModelMixin(SiteModelMixin, models.Model):
    report_datetime = models.DateTimeField(null=True)

    utestid = models.CharField(max_length=25, default="")

    value = models.CharField(max_length=25, default="")

    quantifier = models.CharField(max_length=25, default="")

    value_datetime = models.DateTimeField(null=True)

    reference = models.CharField(max_length=25, default="")

    pending_datetime = models.DateTimeField(default=timezone.now)

    pending = models.BooleanField(default=True)

    resulted_datetime = models.DateTimeField(null=True)

    resulted = models.BooleanField(default=False)

    class Meta:
        abstract = True



# do not delete

from clinicedc_constants import NULL_STRING
from django.db import models


class SingletonPrnModelMixin(models.Model):
    """Enforces one record per subject."""

    singleton_field = models.CharField(
        verbose_name="subject identifier",
        max_length=50,
        unique=True,
        help_text="auto updated for unique constraint",
        default=NULL_STRING,
        editable=False,
    )

    def save(self, *args, **kwargs):
        # enforce singleton constraint on instance, 1 per subject
        self.update_singleton_field()
        super().save(*args, **kwargs)

    def update_singleton_field(self) -> None:
        self.singleton_field = self.subject_identifier

    class Meta:
        abstract = True



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import QuerySet
from django.utils import timezone

if TYPE_CHECKING:
    from ..models import Dispense, DispenseItem, Location, Stock


def dispense(
    stock_codes: list[str],
    location: Location,
    rx,
    dispensed_by: str,
    request: WSGIRequest,
) -> QuerySet[DispenseItem] | None:
    stock_model_cls: type[Stock] = django_apps.get_model("edc_pharmacy.stock")
    dispense_model_cls: type[Dispense] = django_apps.get_model("edc_pharmacy.dispense")
    dispense_item_model_cls: type[DispenseItem] = django_apps.get_model(
        "edc_pharmacy.dispenseitem"
    )

    dispense_cancelled = False
    for stock in stock_model_cls.objects.filter(code__in=stock_codes):
        if stock.allocation.registered_subject.subject_identifier != rx.subject_identifier:
            messages.add_message(
                request,
                messages.ERROR,
                f"Stock not allocated to subject. Got {stock.code}. Dispensing cancelled.",
            )
            dispense_cancelled = True
            break
        if stock.allocation.registered_subject.site.id != stock.location.site.id:
            messages.add_message(
                request,
                messages.ERROR,
                (
                    "Stock location does not match subject's site. "
                    f"Stock item {stock.code} not at site "
                    f"{stock.allocation.registered_subject.site.id}. Dispensing cancelled."
                ),
            )
            dispense_cancelled = True
            break

    if not dispense_cancelled:
        dispense_obj = dispense_model_cls.objects.create(
            rx=rx, location=location, dispensed_by=dispensed_by
        )
        for stock in stock_model_cls.objects.filter(code__in=stock_codes):
            try:
                dispense_item_model_cls.objects.get(stock=stock)
            except ObjectDoesNotExist:
                dispense_item_model_cls.objects.create(
                    dispense=dispense_obj,
                    stock=stock,
                    user_created=request.user.username,
                    created=timezone.now(),
                )
            else:
                messages.add_message(
                    request,
                    messages.ERROR,
                    f"Stock already dispensed. Got {stock.code}.",
                )

        return dispense_item_model_cls.objects.filter(dispense=dispense_obj)
    return None


__all__ = ["dispense"]



from edc_screening.utils import is_eligible_or_raise

from .clean_fields_modelform_validation_mixin import CleanFieldsModelFormValidationMixin
from .consent_modelform_validation_mixin import ConsentModelFormValidationMixin

__all__ = ["ConsentModelFormMixin"]


class ConsentModelFormMixin(
    CleanFieldsModelFormValidationMixin, ConsentModelFormValidationMixin
):
    def clean(self) -> dict:
        cleaned_data = super().clean()
        self.validate_is_eligible_or_raise()
        self.validate_initials_with_full_name()
        self.validate_gender_of_consent()
        self.validate_is_literate_and_witness()
        self.validate_dob_relative_to_consent_datetime()
        self.validate_guardian_and_dob()
        self.validate_identity_and_confirm_identity()
        self.validate_identity_with_unique_fields()
        self.validate_identity_plus_version_is_unique()
        return cleaned_data

    def validate_is_eligible_or_raise(self) -> None:
        screening_identifier = self.get_field_or_raise(
            "screening_identifier", "Screening identifier is required."
        )
        is_eligible_or_raise(screening_identifier=screening_identifier)



from __future__ import annotations

import copy
import sys
from copy import deepcopy
from importlib import import_module
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.management.color import color_style
from django.utils.module_loading import module_has_submodule

if TYPE_CHECKING:
    from edc_prn import Prn


class AlreadyRegistered(Exception):  # noqa: N818
    pass


class SitePrnFormsError(Exception):
    pass


class PrnFormsCollection:
    def __init__(self):
        self.registry: dict[str, Prn] = {}

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __len__(self):
        return len(self.registry.values())

    def __iter__(self):
        return iter(self.registry.values())

    def register(self, prn: Prn = None) -> None:
        if prn.model in self.registry:
            raise AlreadyRegistered(f"PRN form {prn.model} is already registered.")
        self.registry.update({prn.model: prn})
        self.reorder_registry()

    def unregister(self, prn: Prn = None) -> None:
        if prn.model in self.registry:
            del self.registry[prn.model]

    def reorder_registry(self) -> None:
        keys = [k for k in self.registry]
        keys.sort()
        registry = deepcopy(self.registry)
        self.registry = {k: registry.get(k) for k in keys}

    def autodiscover(
        self, module_name: str | None = None, verbose: bool | None = True
    ) -> None:
        module_name = module_name or "prn_forms"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for site {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}                      \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_prn_forms.registry)
                    import_module(f"{app}.{module_name}")
                    writer(f" * registered '{module_name}' from '{app}'\n")
                except SitePrnFormsError as e:
                    writer(f"   - loading {app}.{module_name} ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_prn_forms.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SitePrnFormsError(str(e)) from e
            except ImportError:
                pass


site_prn_forms = PrnFormsCollection()



from django.conf import settings
from django.db import models
from django.db.models import Index, UniqueConstraint
from django.utils.translation import gettext as _

from edc_utils.text import convert_php_dateformat


class Holiday(models.Model):
    id = models.BigAutoField(primary_key=True)

    country = models.CharField(max_length=50)

    local_date = models.DateField()

    name = models.CharField(max_length=50)

    @property
    def label(self) -> str:
        return self.name

    @property
    def formatted_date(self) -> str:
        return self.local_date.strftime(convert_php_dateformat(settings.SHORT_DATE_FORMAT))

    def __str__(self):
        return f"{self.label} on {self.formatted_date}"

    class Meta:
        verbose_name = _("Holiday")
        verbose_name_plural = _("Holidays")
        constraints = [
            UniqueConstraint(
                fields=["country", "local_date"],
                name="%(app_label)s_%(class)s_country_uniq",
            )
        ]
        indexes = (Index(fields=["name", "country", "local_date"]),)



from math import ceil
from urllib.parse import parse_qsl, unquote, urlencode, urljoin

from django import template
from django.conf import settings
from django.template.defaultfilters import stringfilter
from django.urls.base import reverse
from django.utils import timezone
from edc_utils.age import AgeValueError, age

register = template.Library()


class Number:
    def __init__(self, number=None, url=None, current=None):
        self.number = number
        self.url = url
        self.active = "active" if current else ""

    def __str__(self):
        return self.number

    def __repr__(self):
        return f"{self.__class__.__name__}<number={self.number} {self.active}>"


class UrlMaker:
    def __init__(self, base=None, querystring=None):
        self.base = base
        self.querystring = querystring

    def url(self, page):
        url = urljoin(self.base, str(page)) + "/"
        if self.querystring:
            return "?".join([url, self.querystring])
        return url


@register.simple_tag(takes_context=True)
def age_in_years(context, born):
    if born:
        reference_datetime = context.get("reference_datetime") or timezone.now()
        try:
            _age_in_years = age(born, reference_datetime).years
        except AgeValueError:
            _age_in_years = None
    else:
        _age_in_years = context.get("age_in_years")
    return _age_in_years or born


def page_numbers(page, numpages):
    """Returns a list of x integers (display) relative to the value of n
    where n > 0 and the length of the list cannot exceed count.
    """
    page_numbers_ = None
    if page and numpages:
        min_n = page - 5
        min_n = 1 if min_n <= 0 else min_n
        max_n = min_n + 9
        max_n = min(numpages, max_n)
        page_numbers_ = [x for x in range(min_n, max_n + 1)]
    return page_numbers_ or []


@register.inclusion_tag("edc_dashboard/copy_element.html")
def copy_string_to_clipboard_button(value, index=None):
    return dict(value=value, index=index)


@register.inclusion_tag(
    "edc_dashboard/index_link.html",
    takes_context=True,
)
def index_link(context):
    return dict(
        index_page=getattr(settings, "INDEX_PAGE", None),
        index_page_label=getattr(settings, "INDEX_PAGE_LABEL", None),
    )


@register.filter
@stringfilter
def human(value):
    return "-".join([value[i * 4 : (i + 1) * 4] for i in range(0, ceil(len(value) / 4))])


@register.inclusion_tag(
    "edc_dashboard/paginator/paginator_row.html",
    takes_context=True,
)
def paginator_row(context):
    numbers = []
    first_url = None
    previous_url = None
    next_url = None
    last_url = None
    sub_text = None

    paginator_url = context.get("paginator_url")
    paginator = context.get("paginator")
    page_obj = context.get("page_obj")
    querystring = context.get("querystring")

    search_term = context.get("search_term")

    show = page_obj.has_other_pages()

    paginator_url_kwargs = context.get("paginator_url_kwargs") or {}
    if context.get("subject_identifier"):
        paginator_url_kwargs.update(subject_identifier=context.get("subject_identifier"))
    paginator_url = reverse(paginator_url, kwargs=paginator_url_kwargs)

    if querystring:
        if "?" in querystring:
            querystring = querystring.split("?")[1]
        query_dict = parse_qsl(querystring)
        querystring = unquote(urlencode(query_dict))
    if show:
        url_maker = UrlMaker(base=paginator_url, querystring=querystring)
        if page_obj.has_previous():
            first_url = url_maker.url(1)
            previous_url = url_maker.url(page_obj.previous_page_number())
        if page_obj.has_next():
            next_url = url_maker.url(page_obj.next_page_number())
            last_url = url_maker.url(paginator.num_pages)

        for page in page_numbers(page_obj.number, paginator.num_pages):
            current = page_obj.number == page
            url = "#"
            if not current:
                url = url_maker.url(page)
            numbers.append(Number(number=page, url=url, current=current))
        sub_text = (
            f"Showing items {page_obj.start_index()} to {page_obj.end_index()} "
            f"of {paginator.count}."
        )

    return dict(
        paginator_url=paginator_url,
        page_obj=page_obj,
        show=show,
        first_url=first_url,
        previous_url=previous_url,
        next_url=next_url,
        last_url=last_url,
        numbers=numbers,
        search_term=search_term,
        sub_text=sub_text,
    )


@register.filter(name="has_group")
def has_group(user, group_name):
    return user.groups.filter(name=group_name).exists()



from .clinical_review_baseline_form_validator_mixin import (
    ClinicalReviewBaselineFormValidatorMixin,
)
from .clinical_review_followup_form_validator_mixin import (
    ClinicalReviewFollowupFormValidatorMixin,
)



# Generated by Django 5.2.8 on 2025-11-20 22:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_metadata",
            "0033_remove_crfmetadata_edc_metadat_site_id_d3b213_idx_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="crfmetadata",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": (
                    "subject_identifier",
                    "visit_code",
                    "visit_code_sequence",
                    "show_order",
                ),
                "verbose_name": "Crf collection status",
                "verbose_name_plural": "Crf collection status",
            },
        ),
    ]



from .graded_event_notification import GradedEventNotification
from .model_notification import ModelNotification
from .new_model_notification import NewModelNotification
from .notification import Notification
from .updated_model_notification import UpdatedModelNotification



import sys

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.color import color_style

style = color_style()


class RoleUpdaterError(Exception):
    pass


class RoleUpdater:
    def __init__(
        self,
        roles: dict | None = None,
        verbose=None,
    ):
        self.roles = roles
        self.verbose = verbose

    @property
    def role_model_cls(self):
        return django_apps.get_model("edc_auth.role")

    @property
    def group_model_cls(self):
        return django_apps.get_model("auth.group")

    def update_roles(self):
        """Updates or creates role model instances.

        Role model instances refer to group instances. If a group by name
        does not exist, it will be created.
        """
        if self.verbose:
            sys.stdout.write(style.MIGRATE_HEADING(" - Updating roles:\n"))
        for index, (role_name, groups) in enumerate(self.roles.items()):
            display_name = role_name.replace("_", " ").lower().title()
            if self.verbose:
                sys.stdout.write(f"   * updating groups for {role_name}.\n")
            try:
                role = self.role_model_cls.objects.get(name=role_name)
            except ObjectDoesNotExist:
                role = self.role_model_cls.objects.create(
                    name=role_name,
                    display_name=display_name,
                    display_index=index,
                )
            role.display_name = display_name
            role.display_index = index
            role.save()
            role.groups.clear()
            for group_name in groups:
                try:
                    group = self.group_model_cls.objects.get(name=group_name)
                except ObjectDoesNotExist as e:
                    raise RoleUpdaterError(
                        "Invalid group specified for role. "
                        f"`{group_name}` is not a group. See role `{role}`."
                    ) from e
                    # group = self.group_model_cls.objects.create(name=group_name)
                role.groups.add(group)
        if self.verbose:
            sys.stdout.write("   Done.\n")
            sys.stdout.flush()



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import TYPE_CHECKING

from django.contrib.auth import get_permission_codename

from edc_sites.site import sites as site_sites

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django.contrib.sites.models import Site
    from django.db import models

__all__ = ["Perms"]


@dataclass
class SitePerms:
    """Contains add/change/delete/view perms by site.

    Query the sites in the user's profile. Also check
    if the user has additional perms to view sites
    other than the current.

    See also `get_view_only_site_ids_for_user` in edc_sites.sites.
    """

    user: InitVar[User] = None
    current_site: InitVar[Site] = None
    site: InitVar[Site] = None
    add: bool = field(default=False, init=False)
    change: bool = field(default=False, init=False)
    delete: bool = field(default=False, init=False)
    view: bool = field(default=False, init=False)

    def __post_init__(self, user: User, current_site: Site, site: Site) -> None:
        site_sites.site_in_profile_or_raise(user=user, site_id=current_site.id)
        if current_site.id == site.id:
            self.change = True
            self.add = True
            self.view = True
        view_only_sites = site_sites.get_view_only_site_ids_for_user(
            user=user, site_id=current_site.id
        )
        if site.id in view_only_sites:
            # oops, model's site is view only for user
            self.change = False
            self.add = False
            self.view = True


@dataclass
class Perms:
    """Contains model class perms (add/change/delete/view/view_only)
    for this user.

    Here we consider model class, the current site, the sites in
    user's profile, and userprofile.is_multisite_viewer.
    """

    user: User = None
    model_cls: InitVar[type[models.Model]] = None
    current_site: Site = None
    site: Site = None
    add: bool = field(default=False, init=False)
    change: bool = field(default=False, init=False)
    delete: bool = field(default=False, init=False)
    view: bool = field(default=False, init=False)
    view_only: bool = field(default=False, init=False)
    _site_perms: SitePerms = field(default=None, init=False)

    def __post_init__(self, model_cls: type[models.Model]):
        # self.user = get_object_or_404(User, pk=self.user.id)
        # set add, change, delete, view attrs for this user
        # based on the model class
        app_label = model_cls._meta.app_label
        for action in ["add", "change", "view", "delete"]:
            codename = get_permission_codename(action, model_cls._meta)
            setattr(self, action, self.user.has_perm(f"{app_label}.{codename}"))
        # combine the above with permissions relative to the site
        if self.add and not self.site_perms.add:
            self.add = False
        if self.change and not self.site_perms.change:
            self.change = False
        if self.view and not self.site_perms.view:
            self.view = False
        self.view_only = not self.add and not self.change and self.view

    @property
    def site_perms(self) -> SitePerms:
        if not self._site_perms:
            self._site_perms = SitePerms(
                user=self.user,
                current_site=self.current_site,
                site=self.site,
            )
        return self._site_perms



from uuid import uuid4

from clinicedc_constants import NO, NOT_APPLICABLE
from django.core.validators import (
    MaxLengthValidator,
    MaxValueValidator,
    MinLengthValidator,
    MinValueValidator,
    RegexValidator,
)
from django.db import models
from django.utils import timezone
from django_crypto_fields.fields import EncryptedCharField

from edc_constants.choices import GENDER, YES_NO, YES_NO_NA
from edc_model.validators import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_sites.model_mixins import SiteModelMixin


class ScreeningFieldsModeMixin(SiteModelMixin, models.Model):
    reference = models.UUIDField(
        verbose_name="Reference", unique=True, default=uuid4, editable=False
    )

    screening_identifier = models.CharField(
        verbose_name="Screening ID",
        max_length=50,
        blank=True,
        unique=True,
        editable=False,
    )

    report_datetime = models.DateTimeField(
        verbose_name="Report Date and Time",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=timezone.now,
        help_text="Date and time of report.",
    )

    initials = EncryptedCharField(
        validators=[
            RegexValidator("[A-Z]{1,3}", "Invalid format"),
            MinLengthValidator(2),
            MaxLengthValidator(3),
        ],
        help_text="Use UPPERCASE letters only. May be 2 or 3 letters.",
        blank=False,
    )

    gender = models.CharField(choices=GENDER, max_length=10)

    age_in_years = models.IntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(110)]
    )

    ethnicity = models.CharField(max_length=25, default="", blank=True)

    consent_ability = models.CharField(
        verbose_name="Participant or legal guardian/representative able and "
        "willing to give informed consent.",
        max_length=25,
        choices=YES_NO,
    )

    unsuitable_for_study = models.CharField(
        verbose_name=(
            "Is there any other reason the patient is deemed to not be suitable for the study?"
        ),
        max_length=5,
        choices=YES_NO,
        default=NO,
        help_text="If YES, patient NOT eligible, please give reason below.",
    )

    reasons_unsuitable = models.TextField(
        verbose_name="Reason not suitable for the study",
        max_length=150,
        default="",
        blank=True,
    )

    unsuitable_agreed = models.CharField(
        verbose_name=(
            "Does the study coordinator agree that the patient is not suitable for the study?"
        ),
        max_length=5,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    consented = models.BooleanField(default=False, editable=False)

    refused = models.BooleanField(default=False, editable=False)

    class Meta:
        abstract = True



# Generated by Django 5.2.10 on 2026-01-24 15:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0108_alter_container_qty_suggested_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="container",
            old_name="qty_decimal_places",
            new_name="qty_suggested_places",
        ),
        migrations.RenameField(
            model_name="historicalcontainer",
            old_name="qty_decimal_places",
            new_name="qty_suggested_places",
        ),
    ]



from django.contrib import admin
from django.urls import path

from edc_data_manager.views import HomeView
from edc_protocol.research_protocol_config import ResearchProtocolConfig

from .views import SubjectReviewListboardView

app_name = "edc_review_dashboard"

urlpatterns = SubjectReviewListboardView.urls(
    namespace=app_name,
    url_names_key="subject_review_listboard_url",
    identifier_pattern=ResearchProtocolConfig().subject_identifier_pattern,
)

urlpatterns += [
    path("admin/", admin.site.urls),
    path("", HomeView.as_view(), name="home_url"),
]

# aliquot_listboard = ("edc_lab_dashboard:aliquot_listboard_url",)



from typing import Any

from clinicedc_constants import NEW, OPEN, RESOLVED
from django.db.models import QuerySet

from edc_action_item.models import ActionItem
from edc_data_manager.models import DataQuery
from edc_sites.site import sites


class DataManagerViewMixin:
    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(open_data_queries=self.open_data_queries)
        return super().get_context_data(**kwargs)

    @property
    def open_data_queries(self) -> QuerySet:
        """Returns a list of wrapped ActionItem instances
        where status is NEW or OPEN.
        """
        data_queries = DataQuery.objects.filter(
            subject_identifier=self.kwargs.get("subject_identifier"),
        ).exclude(site_response_status=RESOLVED)
        qs = ActionItem.objects.filter(
            pk__in=[obj.action_item.id for obj in data_queries],
            status__in=[NEW, OPEN],
            site_id__in=sites.get_site_ids_for_user(request=self.request),
        )
        return qs



# Generated by Django 4.2.3 on 2023-08-02 23:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_protocol_incident",
            "0018_alter_protocoldeviationviolation_options_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="actionsrequired",
            name="extra_value",
            field=models.CharField(max_length=250, null=True),
        ),
        migrations.AddField(
            model_name="protocolincidents",
            name="extra_value",
            field=models.CharField(max_length=250, null=True),
        ),
        migrations.AddField(
            model_name="protocolviolations",
            name="extra_value",
            field=models.CharField(max_length=250, null=True),
        ),
    ]



from __future__ import annotations

import contextlib

from django.apps import AppConfig

from .list_model_maker import ListModelMaker, ListModelMakerError


class LoadListDataError(Exception):
    pass


def load_list_data(
    list_data: dict | None = None, model_name: str | None = None, apps: AppConfig | None = None
) -> int:
    """Loads data into a list model.

    List models have name, display_name where name
    is the unique field / stored field.

    Format:
        {model_name1: [(name1, display_name),
         (name2, display_name),...],
         model_name2: [(name1, display_name),
         (name2, display_name),...],
        ...}
    """
    model_names = [model_name] if model_name else [k for k in list_data]
    n = 0
    for _model_name in model_names:
        try:
            data = list_data.get(_model_name)()
        except TypeError:
            data = list_data.get(_model_name)
        try:
            for display_index, row in enumerate(data):
                with contextlib.suppress(TypeError):
                    row = row()  # noqa: PLW2901
                maker = ListModelMaker(display_index, row, _model_name, apps=apps)
                maker.create_or_update()
        except ListModelMakerError as e:
            raise LoadListDataError(f"{e} See {list_data.get(_model_name)}.") from e
        n += 1
    return n



from .date import (
    date_is_future,
    date_is_not_now,
    date_is_past,
    date_not_future,
    datetime_is_future,
    datetime_not_future,
)
from .duration import dh_validator, hm_validator, hm_validator2, ymd_validator
from .phone import cell_number, telephone_number



from django.apps import apps as django_apps
from django.db import models
from django.db.models import Index

from edc_model.models import BaseUuidModel, HistoricalRecords


class DataDictionaryManager(models.Manager):
    def get_by_natural_key(self, model, field_name):
        return self.get(model=model, field_name=field_name)


class DataDictionary(BaseUuidModel):
    # see edc_model_to_dataframe
    m2m_related_field = "model"

    model = models.CharField(max_length=250, help_text="label_lower")

    model_verbose_name = models.CharField(max_length=250, null=True)

    number = models.IntegerField()

    prompt = models.CharField(max_length=500)

    field_name = models.CharField(max_length=250)

    field_type = models.CharField(max_length=250, null=True)

    max_length = models.IntegerField(null=True)

    decimal_places = models.IntegerField(null=True)

    max_digits = models.IntegerField(null=True)

    nullable = models.BooleanField(null=True)

    default_value = models.CharField(max_length=250, null=True)

    help_text = models.TextField(null=True)

    active = models.BooleanField(default=False)

    objects = DataDictionaryManager()

    history = HistoricalRecords()

    def save(self, *args, **kwargs):
        self.model_verbose_name = self.get_model_verbose_name()
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.number}. {self.prompt} ({self.model_verbose_name})"

    def natural_key(self):
        return self.model, self.field_name

    def get_model_verbose_name(self):
        verbose_name = self.model
        try:
            model_cls = self.model_cls
        except (LookupError, ValueError):
            pass
        else:
            verbose_name = model_cls._meta.verbose_name
        return verbose_name

    @property
    def model_cls(self):
        return django_apps.get_model(self.model)

    class Meta:
        verbose_name = "Data Dictionary Item"
        verbose_name_plural = "Data Dictionary Items"
        default_permissions = ("view", "export")
        unique_together = (("model", "field_name"),)
        indexes = (Index(fields=["model", "number", "prompt"]),)



from django.core.management import color_style

from edc_utils.dashboard_middleware_check import edc_middleware_check

style = color_style()


def middleware_check(app_configs, **kwargs):
    return edc_middleware_check(
        app_configs,
        app_label="edc_review_dashboard",
        middleware_name="edc_review_dashboard.middleware.DashboardMiddleware",
        error_code=None,
        **kwargs,
    )



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django import forms

from edc_appointment.constants import MISSED_APPT
from edc_sites.modelform_mixins import SiteModelFormMixin
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit_schedule import VisitSchedule

from ..constants import MISSED_VISIT

if TYPE_CHECKING:
    from django.contrib.sites.models import Site

    from edc_appointment.models import Appointment


class VisitTrackingModelFormMixin(SiteModelFormMixin):
    report_datetime_field_attr = "report_datetime"

    def clean(self):
        cleaned_data = super().clean()
        if (
            cleaned_data.get("reason")
            and self.appointment.appt_timing != MISSED_APPT
            and cleaned_data.get("reason") == MISSED_VISIT
        ):
            raise forms.ValidationError(
                {
                    "reason": (
                        "Invalid. Appointment is missed. Expected visit to be missed also."
                    )
                }
            )
        if (
            cleaned_data.get("reason")
            and self.appointment.appt_timing == MISSED_APPT
            and cleaned_data.get("reason") != MISSED_VISIT
        ):
            raise forms.ValidationError(
                {
                    "reason": (
                        "Invalid. Appointment is not missed. Did not expected a missed visit."
                    )
                }
            )
        return cleaned_data

    @property
    def subject_identifier(self) -> str:
        return self.get_subject_identifier()

    def get_subject_identifier(self) -> str:
        return self.appointment.subject_identifier

    @property
    def report_datetime(self) -> datetime:
        return self.cleaned_data.get(self.report_datetime_field_attr) or getattr(
            self.instance, self.report_datetime_field_attr
        )

    @property
    def appointment(self) -> Appointment:
        return self.cleaned_data.get("appointment") or self.instance.appointment

    @property
    def visit_schedule_name(self) -> str:
        return self.visit_schedule.name

    @property
    def schedule_name(self) -> str:
        return self.schedule.name

    @property
    def visit_schedule(self) -> VisitSchedule:
        return self.appointment.visit_schedule

    @property
    def schedule(self) -> Schedule:
        return self.appointment.schedule

    @property
    def site(self) -> Site:
        return self.cleaned_data.get("site") or self.appointment.site



requisition_fieldset: tuple[str, dict[str, tuple[str, ...]]] = (
    "Requisition",
    {
        "fields": (
            "is_drawn",
            "reason_not_drawn",
            "reason_not_drawn_other",
            "drawn_datetime",
            "item_type",
            "item_count",
            "estimated_volume",
            "comments",
        )
    },
)


requisition_status_fields: tuple[str, ...] = (
    "received",
    "received_datetime",
    "processed",
    "processed_datetime",
    "packed",
    "packed_datetime",
    "shipped",
    "shipped_datetime",
)

requisition_verify_fields: tuple[str, ...] = (
    "clinic_verified",
    "clinic_verified_datetime",
)

requisition_status_fieldset: tuple[str, dict[str, tuple[str, ...]]] = (
    "Status (For laboratory use only)",
    {"classes": ("collapse",), "fields": requisition_status_fields},
)


requisition_identifier_fields: tuple[str, ...] = (
    "requisition_identifier",
    "identifier_prefix",
    "primary_aliquot_identifier",
)

requisition_identifier_fieldset: tuple[str, dict[str, tuple[str, ...]]] = (
    "Identifiers",
    {"classes": ("collapse",), "fields": requisition_identifier_fields},
)

requisition_verify_fieldset: tuple[str, dict[str, tuple[str, ...]]] = (
    "Verification",
    {"classes": ("collapse",), "fields": requisition_verify_fields},
)



from __future__ import annotations

from django.utils.translation import gettext_lazy as _


class TemplatesModelAdminMixin:
    """Override admin templates.

    Note: If using inlines.

    On the inline admin class specify the position `after` with class
    attribute `insert_after`:

    For example:
        class MyInlineModelAdmin(..):
            ...
            insert_after="<fieldname>"
            ...

    See also: https://linevi.ch/en/django-inline-in-fieldset.html
    """

    show_object_tools: bool = False
    view_on_site_label: str = _("View on site")
    history_label: str = _("Audit trail")
    show_history_label: bool = True

    add_form_template: str = "edc_model_admin/admin/change_form.html"
    change_form_template: str = "edc_model_admin/admin/change_form.html"
    change_list_template: str = "edc_model_admin/admin/change_list.html"
    change_list_title: str | None = None
    change_list_note: str | None = None
    change_list_help: str | None = None
    delete_confirmation_template: str = "edc_model_admin/admin/delete_confirmation.html"
    change_form_title: str | None = None

    def rendered_change_list_note(self):
        """Override if the class attr `change_list_note` is more than
        a normal string.

        For example, if the rendering attempts to resolve a URL for
        a class not yet registered, modeladmin registration silently
        stops registering the remaining modeladmin classes.
        """
        return self.change_list_note

    def change_view(self, request, object_id, form_url="", extra_context=None):
        extra_context = extra_context or {}
        extra_context.update(
            {
                "view_on_site_label": self.view_on_site_label,
                "show_history_label": self.show_history_label,
                "history_label": self.history_label,
                "title": self.change_form_title,
            }
        )
        return super().change_view(
            request, object_id, form_url=form_url, extra_context=extra_context
        )

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context if extra_context else {}
        extra_context.update(
            {
                "show_object_tools": self.show_object_tools,
                "change_list_note": self.rendered_change_list_note,
                "change_list_help": self.change_list_help,
            }
        )
        if self.change_list_title:
            extra_context.update({"title": self.change_list_title.title()})
        return super().changelist_view(request, extra_context=extra_context)



from __future__ import annotations

import re
from datetime import date, datetime

from django.db import models

from edc_model.models import BaseUuidModel, HistoricalRecords

from ..exceptions import ValueBoundryError
from .reference_model_mixins import ReferenceModelMixin


class NormalData(ReferenceModelMixin, BaseUuidModel):
    auto_created = models.BooleanField(default=False)

    history = HistoricalRecords()

    def __str__(self):
        return self.description

    def value_in_normal_range_or_raise(
        self,
        value: int | float,
        dob: date,
        report_datetime: datetime,
        age_units: str | None = None,
    ) -> bool:
        """Raises a ValueBoundryError exception if condition not met.

        Condition is evaluated to True or False as a string
        constructed from given parameters.

        For example,
            "lower lower_operator value upper_operator upper"
            "1.7<3.6<=3.5"
            "7.3<3.6"
        """

        pattern = r"([<>]=?|==|!=)?\s*-?\d+(\.\d+)?"

        self.age_in_bounds_or_raise(dob, report_datetime, age_units)

        value = float(value)
        value_condition_str = self.get_eval_phrase(value)
        if not re.match(pattern, value_condition_str):
            raise ValueError(f"Invalid condition string. Got {value_condition_str}.")
        if not eval(value_condition_str):  # nosec B307  # noqa: S307
            raise ValueBoundryError(
                f"{self.label}: {value_condition_str}{self.units} [{self.gender}]"
            )
        return True

    def get_eval_phrase(self, value) -> str:
        return (
            f"{'' if not self.lower else self.lower}{self.lower_operator or ''}{value}"
            f"{self.upper_operator or ''}{'' if not self.upper else self.upper}"
        )

    class Meta:
        verbose_name = "Normal Reference"
        verbose_name_plural = "Normal References"



from .add_or_update_django_sites import add_or_update_django_sites
from .get_message_text import get_message_text
from .get_or_create_site_obj import get_or_create_site_obj
from .get_or_create_site_profile_obj import get_or_create_site_profile_obj
from .get_site_model_cls import get_site_model_cls
from .has_profile_or_raise import has_profile_or_raise
from .insert_into_domain import insert_into_domain
from .valid_site_for_subject_or_raise import valid_site_for_subject_or_raise



from django.contrib import admin
from django_audit_fields import audit_fieldset_tuple

from edc_model_admin.mixins import TemplatesModelAdminMixin

from ..admin_site import edc_reportable_admin
from ..models import NormalData


@admin.register(NormalData, site=edc_reportable_admin)
class NormalDataAdmin(TemplatesModelAdminMixin, admin.ModelAdmin):

    fieldsets = (
        (
            None,
            {
                "fields": [
                    "label",
                    "units",
                    "gender",
                    "lower",
                    "lower_operator",
                    "lln",
                    "upper",
                    "upper_operator",
                    "uln",
                    "age_lower",
                    "age_lower_operator",
                    "age_upper",
                    "age_upper_operator",
                    "reference_range_collection",
                ]
            },
        ),
        audit_fieldset_tuple,
    )

    ordering = ("label", "units", "gender")
    list_display = (
        "label",
        "units",
        "gender",
        "range_desc",
        "age_desc",
        "collection",
        "auto_created",
        "created",
        "modified",
    )

    list_filter = (
        "reference_range_collection__name",
        "gender",
        "auto_created",
        "label",
        "units",
    )

    search_fields = ("label",)

    @admin.display(description="Range", ordering="lower")
    def range_desc(self, obj: NormalData) -> str | None:
        if obj and (obj.lower or obj.upper):
            return (
                f"{obj.lower or ''}{obj.lln or ''}{obj.lower_operator or ''}x"
                f"{obj.upper_operator or ''}{obj.upper or ''}{obj.uln or ''}"
            )
        return None

    @admin.display(description="Age", ordering="age_lower")
    def age_desc(self, obj: NormalData) -> str | None:
        if obj and obj.age_phrase:
            return obj.age_phrase % dict(age_value="x")
        return None

    @admin.display(description="Collection", ordering="collection__name")
    def collection(self, obj: NormalData) -> str | None:
        return obj.reference_range_collection.name



# Generated by Django 3.2.11 on 2022-04-12 20:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_pharmacy", "0002_alter_medication_unique_together"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="rxrefill",
            unique_together={
                ("rx", "refill_date"),
                ("rx", "visit_code", "visit_code_sequence"),
            },
        ),
    ]



BLOOD_RESULTS_EGFR_ACTION = "abnormal-blood-results-egfr"
BLOOD_RESULTS_FBC_ACTION = "abnormal-blood-results-fbc"
BLOOD_RESULTS_GLU_ACTION = "abnormal-blood-results-glu"
BLOOD_RESULTS_HBA1C_ACTION = "abnormal-blood-results-hba1c"
BLOOD_RESULTS_INSULIN_ACTION = "abnormal-blood-results-ins"
BLOOD_RESULTS_LFT_ACTION = "abnormal-blood-results-lft"
BLOOD_RESULTS_LIPIDS_ACTION = "abnormal-blood-results-lipid"
BLOOD_RESULTS_RFT_ACTION = "abnormal-blood-results-rft"
URINALYSIS_ACTION = "abnormal-urinalysis"



import socket
from typing import Any

from django.apps import apps as django_apps


def get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    try:
        socket.inet_aton(ip)
    except OSError:
        return None
    else:
        return ip


class EdcDeviceViewMixin:
    def get_context_data(self, **kwargs) -> dict[str, Any]:
        app_config = django_apps.get_app_config("edc_device")
        kwargs.update(
            {
                "device_id": app_config.device_id,
                "device_role": app_config.device_role,
                "ip_address": self.ip_address,
            }
        )
        return super().get_context_data(**kwargs)

    @property
    def ip_address(self):
        client_ip = None
        if self.request:
            client_ip = get_client_ip(self.request)
        return client_ip



from django.db import models
from django.utils import timezone

from ..screening_eligibility import ScreeningEligibility


class EligibilityFieldsModelMixin(models.Model):
    eligible = models.BooleanField(default=False)

    reasons_ineligible = models.TextField(
        verbose_name="Reason not eligible", max_length=150, default=""
    )

    eligibility_datetime = models.DateTimeField(
        null=True,
        help_text="Date and time eligibility was determined relative to report_datetime",
    )

    real_eligibility_datetime = models.DateTimeField(
        null=True,
        help_text="Date and time eligibility was determined relative to now",
    )

    def get_report_datetime_for_eligibility_datetime(self):
        """Returns report_datetime.

        Override to use a different report_datetime if, for example,
        screening is done in more than one part.
        """
        return self.report_datetime

    class Meta:
        abstract = True


class EligibilityModelMixin(EligibilityFieldsModelMixin, models.Model):
    eligibility_cls = ScreeningEligibility

    def save(self, *args, **kwargs):
        """When saved, the eligibility_cls is instantiated and the
        value of `eligible` is evaluated.

        * If not eligible, updates reasons_ineligible.
        * Screening Identifier is always allocated.
        """
        # if self.eligibility_cls:
        self.eligibility_cls(model_obj=self)
        if not self.id:
            self.screening_identifier = self.identifier_cls().identifier
        if self.eligible:
            self.eligibility_datetime = self.get_report_datetime_for_eligibility_datetime()
            self.real_eligibility_datetime = timezone.now()
        else:
            self.eligibility_datetime = None
            self.real_eligibility_datetime = None
        super().save(*args, **kwargs)

    class Meta:
        abstract = True



OFFSTUDY = "OFFSTUDY"
OFFSTUDY_VIEW = "OFFSTUDY_VIEW"
OFFSTUDY_SUPER = "OFFSTUDY_SUPER"
codenames = [
    "edc_offstudy.add_subjectoffstudy",
    "edc_offstudy.change_subjectoffstudy",
    "edc_offstudy.delete_subjectoffstudy",
    "edc_offstudy.view_historicalsubjectoffstudy",
    "edc_offstudy.view_subjectoffstudy",
]



import contextlib
import sys
from warnings import warn

from django.apps import apps as django_apps
from django.conf import settings
from django.db import OperationalError, connection
from edc_auth.get_app_codenames import get_app_codenames


def get_qareports_codenames(app_name: str, *note_models: str) -> list[str]:
    warn(
        "This function has been deprecated. Use get_app_codenames.",
        DeprecationWarning,
        2,
    )
    return get_app_codenames(app_name)


def recreate_db_view(model_cls, drop: bool | None = None, verbose: bool | None = None):
    """Manually recreate the database view for models declared
    with `django_db_views.DBView`.

    Mostly useful when Django raises an OperationalError with a
    restored DB complaining of 'The user specified as a definer
    (user@host) does not exist' or some variation of OperationalError
    with 'SELECT command denied to user...'

    This does not replace generating a migration with `viewmigration`
    and running the migration.

    For example:
        from intecomm_reports.models import Vl

        Vl.recreate_db_view()

    Also, could do something like this (replace details as required):
        CREATE USER 'edc-effect-live'@'10.131.23.168' IDENTIFIED BY 'xxxxxx';
        GRANT SELECT ON effect_prod.* to 'edc-effect-live'@'10.131.23.168';

    You can run through all models using this mixin and recreate:

        from django.apps import apps as django_apps
        from edc_qareports.model_mixins import QaReportModelMixin

        for model_cls in django_apps.get_models():
            if issubclass(model_cls, (QaReportModelMixin,)):
                print(model_cls)
                try:
                    model_cls.recreate_db_view()
                except AttributeError as e:
                    print(e)
                except TypeError as e:
                    print(e)
    """
    drop = True if drop is None else drop
    try:
        sql = model_cls.view_definition.get(settings.DATABASES["default"]["ENGINE"])
    except AttributeError as e:
        raise AttributeError(
            f"Is this model linked to a view? Declare model with `DBView`. Got {e}"
        ) from e
    else:
        sql = sql.replace(";", "")
        if verbose:
            sys.stdout.write(f"create view {model_cls._meta.db_table} as {sql};\n")
        with connection.cursor() as c:
            if drop:
                with contextlib.suppress(OperationalError):
                    c.execute(f"drop view {model_cls._meta.db_table};")
            c.execute(f"create view {model_cls._meta.db_table} as {sql};")
        if verbose:
            sys.stdout.write(
                f"Done. Refreshed DB VIEW `{model_cls._meta.db_table}` for model {model_cls}."
            )


def recreate_dbview_for_all():
    from .model_mixins import QaReportModelMixin  # noqa: PLC0415

    for model_cls in django_apps.get_models():
        if issubclass(model_cls, (QaReportModelMixin,)):
            sys.stdout.write(f"{model_cls}\n")
            try:
                model_cls.recreate_db_view()
            except AttributeError as e:
                sys.stdout.write(f"{e}\n")
            except TypeError as e:
                sys.stdout.write(f"{e}\n")



from copy import deepcopy

from clinicedc_constants import YES

from edc_form_validators import INVALID_ERROR
from edc_registration import get_registered_subject_model_cls

__all__ = ["ReportablesFormValidatorMixin"]

from ..exceptions import NotEvaluated
from ..reference_range_evaluator import ReferenceRangeEvaluator


class ReportablesFormValidatorMixin:
    value_field_suffix = "_value"
    reference_range_evaluator_cls = ReferenceRangeEvaluator

    @property
    def age_units(self):
        return "years"

    @property
    def reportables_evaluator_options(self):
        return {"age_units": "years"}

    def validate_reportable_fields(
        self,
        reference_range_collection_name: str,
        age_units: str | None = None,
        **reportables_evaluator_options,
    ):
        """Called in clean() method of the FormValidator.

        for example:

            def clean(self):
                ...
                self.validate_reportable_fields()
                ...
        """

        cleaned_data = deepcopy(self.cleaned_data)
        registered_subject = get_registered_subject_model_cls().objects.get(
            subject_identifier=self.subject_identifier
        )
        # check normal ranges and grade result values
        options = dict(
            cleaned_data=deepcopy(cleaned_data),
            gender=registered_subject.gender,
            dob=registered_subject.dob,
            report_datetime=self.report_datetime,
            value_field_suffix=self.value_field_suffix,
        )
        age_units = age_units or self.age_units
        options.update(**reportables_evaluator_options)
        reference_range_evaluator = self.reference_range_evaluator_cls(
            reference_range_collection_name, age_units=age_units, **options
        )
        try:
            reference_range_evaluator.validate_reportable_fields()
        except NotEvaluated as e:
            self.raise_validation_error({"__all__": str(e)}, INVALID_ERROR, exc=e)
        reference_range_evaluator.validate_results_abnormal_field()
        self.applicable_if(
            YES, field="results_abnormal", field_applicable="results_reportable"
        )
        reference_range_evaluator.validate_results_reportable_field()



from __future__ import annotations

from datetime import datetime

from django import forms

from ..subject_schedule import InvalidOffscheduleDate
from .visit_schedule_non_crf_modelform_mixin import VisitScheduleNonCrfModelFormMixin


class OffScheduleModelFormMixin(VisitScheduleNonCrfModelFormMixin):
    offschedule_datetime_field_attr = "offschedule_datetime"

    def clean(self):
        cleaned_data = super().clean()
        if self.offschedule_datetime:
            history_obj = self.schedule.history_model_cls.objects.get(
                subject_identifier=self.get_subject_identifier(),
                schedule_name=self.schedule_name,
                visit_schedule_name=self.visit_schedule_name,
            )
            try:
                self.schedule.subject(self.get_subject_identifier()).update_history_or_raise(
                    history_obj=history_obj,
                    offschedule_datetime=self.offschedule_datetime,
                    update=False,
                )
            except InvalidOffscheduleDate as e:
                raise forms.ValidationError(e) from e
            self.validate_visit_tracking_reports()
        return cleaned_data

    # TODO: validate_visit_tracking_reports before taking off schedule
    def validate_visit_tracking_reports(self):
        """Asserts that all visit tracking reports
        have been submitted.
        """
        pass

    @property
    def offschedule_datetime(self) -> datetime | None:
        if self.offschedule_datetime_field_attr in self.cleaned_data:
            return self.cleaned_data.get(self.offschedule_datetime_field_attr)
        return getattr(self.instance, self.offschedule_datetime_field_attr)

    @property
    def offschedule_compare_dates_as_datetimes(self):
        return True

    class Meta:
        help_text = {  # noqa: RUF012
            "subject_identifier": "(read-only)",
            "action_identifier": "(read-only)",
        }
        widgets = {  # noqa: RUF012
            "subject_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
            "action_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from django import template

register = template.Library()


@register.inclusion_tag("edc_label/print_button.html")
def print_button(button_label=None, printer=None, label_template_name=None, **kwargs):
    return dict(
        button_label=button_label,
        label_template_name=label_template_name,
        printer=printer,
    )


@register.inclusion_tag("edc_label/printer_config.html")
def printer_config(
    heading=None, printer_type=None, selected_printer=None, printers=None, **kwargs
):
    return dict(
        heading=heading,
        printer_type=printer_type,
        selected_printer=selected_printer,
        printers=printers,
    )



from django.apps import apps as django_apps
from django.conf import settings
from django.db.models.signals import pre_save
from django.dispatch import receiver

from .device_permission import device_permissions


def update_device_fields(instance):
    device_id = getattr(settings, "DEVICE_ID", None)
    app_config = django_apps.get_app_config("edc_device")
    if not instance.id:
        instance.device_created = device_id or app_config.device_id
    instance.device_modified = device_id or app_config.device_id


@receiver(pre_save, weak=False, dispatch_uid="check_device_on_pre_save")
def check_device_on_pre_save(sender, instance, raw, using, update_fields, **kwargs):
    """Updates device id."""
    if not raw:
        try:
            instance.device_created
            instance.device_modified
        except AttributeError:
            pass
        else:
            update_device_fields(instance)
            device_permissions.check(instance)



from __future__ import annotations

from datetime import datetime
from operator import methodcaller
from typing import TYPE_CHECKING, Any
from zoneinfo import ZoneInfo

import arrow
from arrow import Arrow
from dateutil.relativedelta import relativedelta, weekday
from django.conf import settings
from django.utils import timezone

from edc_utils.text import convert_php_dateformat

from .exceptions import FacilityError
from .holidays import Holidays

if TYPE_CHECKING:
    pass


class Facility:
    """
    Note: `best_effort_available_datetime` (Default: False) if True
        will set available_rdata to the suggested_datetime if no
        available_rdata can be found. This is not ideal and could
        lead to a protocol violation but may be helpful for facilities
        open 1 or 2 days per week, where the visit has a very
        narrow window period (forward_delta, reverse_delta).
    """

    holiday_cls = Holidays

    def __init__(
        self,
        name: str | None = None,
        days: list[weekday] | None = None,
        slots: list[int] | None = None,
        best_effort_available_datetime: datetime | None = None,
    ):
        self.days = days
        self.name = name
        if not name:
            raise FacilityError(f"Name cannot be None. See {self!r}")
        self.best_effort_available_datetime = (
            True if best_effort_available_datetime is None else best_effort_available_datetime
        )
        self.weekdays = [d.weekday for d in self.days]
        self.slots = slots or [99999 for _ in self.days]
        self.config = dict(zip([str(d) for d in self.days], self.slots, strict=False))
        self.holidays = self.holiday_cls()

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name}, days={self.days})"

    def __str__(self):
        description = ", ".join(
            [str(day) + "(" + str(slot) + " slots)" for day, slot in self.config.items()]
        )
        return f"{self.name.title()} {description}"

    def slots_per_day(self, day) -> int:
        try:
            slots_per_day = self.config.get(str(day))
        except KeyError:
            slots_per_day = 0
        return slots_per_day

    @staticmethod
    def open_slot_on(arr) -> Arrow:
        """Hook for handling load balance by day.

        For example, if 15 appointment `slots` are filled out of 30
        allowed for Monday, return arw, if 30/30 return None.
        """
        return arr

    def is_holiday(self, dte: datetime) -> bool:
        return self.holidays.is_holiday(dte=dte)

    def available_datetime(self, **kwargs) -> datetime:
        return self.available_arr(**kwargs).datetime

    @staticmethod
    def get_arr_span(
        suggested_arr, forward_delta, reverse_delta
    ) -> tuple[list[Arrow | Any], Arrow, Arrow]:
        """Returns a list of arrow objects in a custom ordered.
        Objects are ordered around the suggested date. For example,
        """
        # min_arw = self.to_arrow_utc(suggested_arr.datetime - reverse_delta)
        min_arr = Arrow.fromdate(
            suggested_arr.datetime - reverse_delta, tzinfo=ZoneInfo(settings.TIME_ZONE)
        )
        # max_arw = self.to_arrow_utc(suggested_arw.datetime + forward_delta)
        max_arr = Arrow.fromdate(
            suggested_arr.datetime + forward_delta, tzinfo=ZoneInfo(settings.TIME_ZONE)
        )
        span = [arw[0] for arw in Arrow.span_range("day", min_arr.datetime, max_arr.datetime)]
        span_lt = [arw for arw in span if arw.date() < suggested_arr.date()]
        span_lt = sorted(span_lt, key=methodcaller("date"), reverse=True)
        span_gt = [arw for arw in span if arw.date() > suggested_arr.date()]
        arr_span = []
        max_len = max(len(span_lt), len(span_gt))
        for _ in range(0, max_len):
            try:
                item = span_lt.pop()
            except IndexError:
                pass
            else:
                arr_span.insert(0, item)
            try:
                item = span_gt.pop()
            except IndexError:
                pass
            else:
                arr_span.insert(0, item)
        arr_span.insert(0, suggested_arr)
        return arr_span, min_arr, max_arr

    def available_arr(
        self,
        suggested_datetime=None,
        forward_delta=None,
        reverse_delta=None,
        taken_datetimes=None,
        schedule_on_holidays=None,
        **kwargs,  # noqa: ARG002
    ):
        """Returns an arrow object for a datetime equal to or
        close to the suggested datetime.

        To exclude datetimes other than holidays, pass a list of
        datetimes to `taken_datetimes`.
        """
        available_arr = None
        forward_delta = forward_delta or relativedelta(months=1)
        reverse_delta = reverse_delta or relativedelta(months=0)
        taken_arr = [
            arrow.Arrow.fromdatetime(dt, tzinfo=ZoneInfo(settings.TIME_ZONE))
            for dt in taken_datetimes or []
        ]
        if suggested_datetime:
            suggested_arr = arrow.Arrow.fromdatetime(suggested_datetime)
        else:
            suggested_arr = arrow.Arrow.fromdatetime(timezone.now())
        arr_span_range, min_arr, max_arr = self.get_arr_span(
            suggested_arr,
            forward_delta,
            reverse_delta,
        )
        for arr in arr_span_range:
            # add back time to arrow object, r
            if arr.date().weekday() in self.weekdays and (
                min_arr.date() <= arr.date() < max_arr.date()
            ):
                is_holiday = False if schedule_on_holidays else self.is_holiday(arr.datetime)
                if (
                    not is_holiday
                    and arr.date() not in [a.date() for a in taken_arr]
                    and self.open_slot_on(arr)
                ):
                    available_arr = arr
                    break
        if not available_arr:
            if self.best_effort_available_datetime:
                available_arr = suggested_arr
            else:
                formatted_date = suggested_datetime.strftime(
                    convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                )
                raise FacilityError(
                    f"No available appointment dates at facility for period. "
                    f"Got no available dates within {reverse_delta.days}-"
                    f"{forward_delta.days} days of {formatted_date}. "
                    f"Facility is {self!r}."
                )
        return arrow.Arrow.fromdatetime(
            datetime.combine(available_arr.date(), suggested_arr.time())
        )



from django.db import models

from edc_constants.choices import YES_NO
from edc_glucose.model_mixins import Hba1cModelMixin as BaseHba1cModelMixin
from edc_reportable.choices import REPORTABLE


class Hba1cModelMixin(BaseHba1cModelMixin, models.Model):
    is_poc = models.CharField(
        verbose_name="Was a point-of-care test used?",
        max_length=15,
        choices=YES_NO,
        default="",
    )

    hba1c_abnormal = models.CharField(
        verbose_name="abnormal", choices=YES_NO, max_length=25, default="", blank=True
    )

    hba1c_reportable = models.CharField(
        verbose_name="reportable",
        choices=REPORTABLE,
        max_length=25,
        default="",
        blank=True,
    )

    class Meta:
        abstract = True



from clinicedc_constants import EQ, NOT_APPLICABLE
from django.db import models

from edc_lab.choices import GLUCOSE_UNITS_NA, RESULT_QUANTIFIER

from ..constants import GLUCOSE_HIGH_READING


def fbg_model_mixin_factory(utest_id: str, **kwargs):
    class AbstractModel(models.Model):
        class Meta:
            abstract = True

    opts = {
        f"{utest_id}_value": models.DecimalField(
            verbose_name="FBG level",
            max_digits=8,
            decimal_places=2,
            null=True,
            blank=True,
            help_text=f"A `HIGH` reading may be entered as {GLUCOSE_HIGH_READING}",
        ),
        f"{utest_id}_quantifier": models.CharField(
            verbose_name="FBG quantifier",
            max_length=10,
            choices=RESULT_QUANTIFIER,
            default=EQ,
        ),
        f"{utest_id}_units": models.CharField(
            verbose_name="FBG units",
            max_length=15,
            choices=GLUCOSE_UNITS_NA,
            default=NOT_APPLICABLE,
        ),
        f"{utest_id}_datetime": models.DateTimeField(
            verbose_name="FBG date/time measured",
            null=True,
            blank=True,
        ),
    }

    opts.update(**kwargs)

    for name, fld_cls in opts.items():
        AbstractModel.add_to_class(name, fld_cls)

    return AbstractModel



# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0046_infosources"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="appointment",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="appointmenttype",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Appointment Type",
                "verbose_name_plural": "Appointment Types",
            },
        ),
        migrations.AlterModelOptions(
            name="infosources",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Information Source",
                "verbose_name_plural": "Information Sources",
            },
        ),
        migrations.RemoveIndex(
            model_name="appointment",
            name="edc_appoint_subject_56a935_idx",
        ),
        migrations.RemoveIndex(
            model_name="infosources",
            name="edc_appoint_id_ffb4c6_idx",
        ),
        migrations.AlterUniqueTogether(
            name="appointment",
            unique_together=set(),
        ),
        migrations.AddField(
            model_name="appointment",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="appointment",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="appointmenttype",
            name="display_index",
            field=models.IntegerField(
                default=0,
                help_text="Index to control display order if not alphabetical, not required",
                verbose_name="display index",
            ),
        ),
        migrations.AlterField(
            model_name="appointmenttype",
            name="display_name",
            field=models.CharField(
                help_text="(suggest 40 characters max.)",
                max_length=250,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="appointmenttype",
            name="name",
            field=models.CharField(
                help_text="This is the stored value, required",
                max_length=250,
                unique=True,
                verbose_name="Stored value",
            ),
        ),
        migrations.AlterField(
            model_name="infosources",
            name="display_index",
            field=models.IntegerField(
                default=0,
                help_text="Index to control display order if not alphabetical, not required",
                verbose_name="display index",
            ),
        ),
        migrations.AlterField(
            model_name="infosources",
            name="display_name",
            field=models.CharField(
                help_text="(suggest 40 characters max.)",
                max_length=250,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="infosources",
            name="name",
            field=models.CharField(
                help_text="This is the stored value, required",
                max_length=250,
                unique=True,
                verbose_name="Stored value",
            ),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(
                fields=["appt_datetime"], name="edc_appoint_appt_da_7c2e7a_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(fields=["appt_status"], name="edc_appoint_appt_st_57c8d0_idx"),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(
                fields=["timepoint", "visit_code_sequence"],
                name="edc_appoint_timepoi_c6dbd3_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(
                fields=["subject_identifier", "appt_reason"],
                name="edc_appoint_subject_e1b995_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(
                fields=[
                    "site",
                    "subject_identifier",
                    "timepoint",
                    "visit_code_sequence",
                ],
                name="edc_appoint_site_id_312976_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "appt_reason",
                    "timepoint",
                    "visit_code_sequence",
                ],
                name="edc_appoint_subject_f115dc_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(
                fields=["modified", "created"], name="edc_appoint_modifie_fa1fe2_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="appointment",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_appoint_user_mo_9cfb72_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="appointmenttype",
            index=models.Index(
                fields=["display_index", "display_name"],
                name="edc_appoint_display_386d1e_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="infosources",
            index=models.Index(
                fields=["display_index", "display_name"],
                name="edc_appoint_display_38737a_idx",
            ),
        ),
        migrations.AddConstraint(
            model_name="appointment",
            constraint=models.UniqueConstraint(
                fields=(
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "timepoint",
                    "visit_code_sequence",
                ),
                name="unique_edc_appointment_appointment_100",
            ),
        ),
        migrations.AddConstraint(
            model_name="appointment",
            constraint=models.UniqueConstraint(
                fields=(
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "appt_datetime",
                ),
                name="unique_edc_appointment_appointment_200",
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-03-31 12:32

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0024_alter_manifestitem_managers"),
    ]

    operations = [
        migrations.AlterField(
            model_name="aliquot",
            name="measure_units",
            field=models.CharField(
                choices=[
                    ("mL", "mL"),
                    ("uL", "uL"),
                    ("spots", "spots"),
                    ("n/a", "Not applicable"),
                ],
                default="mL",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="measure_units",
            field=models.CharField(
                choices=[
                    ("mL", "mL"),
                    ("uL", "uL"),
                    ("spots", "spots"),
                    ("n/a", "Not applicable"),
                ],
                default="mL",
                max_length=25,
            ),
        ),
    ]



from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone
from edc_model.models import BaseUuidModel, HistoricalRecords
from sequences import get_next_value

from ...exceptions import ReceiveError
from .location import Location
from .order import Order
from .supplier import Supplier


class Manager(models.Manager):
    use_in_migrations = True


class Receive(BaseUuidModel):
    receive_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    receive_datetime = models.DateTimeField(default=timezone.now)

    item_count = models.IntegerField(
        verbose_name="Item count",
        null=True,
        validators=[MinValueValidator(1)],
    )

    location = models.ForeignKey(
        Location,
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        limit_choices_to={"site__isnull": True},
    )

    order = models.ForeignKey(Order, on_delete=models.PROTECT, null=True, blank=False)

    supplier = models.ForeignKey(
        Supplier,
        on_delete=models.PROTECT,
        verbose_name="Supplier",
        null=True,
        blank=False,
    )

    invoice_number = models.CharField(max_length=50, default="", blank=True)

    invoice_date = models.DateField(null=True, blank=True)

    comment = models.TextField(default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.receive_identifier

    def save(self, *args, **kwargs):
        if not self.receive_identifier:
            self.receive_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        if not self.order:
            raise ReceiveError("Order may not be null.")
        if not self.location:
            raise ReceiveError("Location may not be null.")
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Receive"
        verbose_name_plural = "Receive"



from __future__ import annotations

import ast

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Count
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin

from ..exceptions import AllocationError, InsufficientStockError
from ..models import Assignment, Stock, StockRequest, StockRequestItem
from ..utils import allocate_stock


@method_decorator(login_required, name="dispatch")
class AllocateToSubjectView(EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView):
    model_pks: list[str] | None = None
    template_name: str = "edc_pharmacy/stock/allocate_to_subject.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"
    items_per_page = 12

    def get_context_data(self, **kwargs):
        remaining_count, total_count = self.get_counts(self.stock_request)
        show_count = min(self.items_per_page, remaining_count)
        kwargs.update(
            stock_request=self.stock_request,
            assignment=self.selected_assignment,
            stock_request_changelist_url=self.stock_request_changelist_url,
            subject_identifiers=self.get_next_subject_identifiers(self.items_per_page),
            subject_identifiers_count=self.subject_identifiers.count(),
            assignments=Assignment.objects.all().order_by("name"),
            remaining_count=remaining_count,
            total_count=total_count,
            show_count=show_count,
            SHORT_DATE_FORMAT=settings.SHORT_DATE_FORMAT,
        )
        return super().get_context_data(**kwargs)

    @property
    def subject_identifiers(self):
        """Returns a queryset of unallocated stock request
        items for the given assignment.
        """
        return (
            StockRequestItem.objects.values_list(
                "registered_subject__subject_identifier", flat=True
            )
            .filter(
                stock_request=self.stock_request,
                allocation__isnull=True,
                assignment=self.selected_assignment,
            )
            .order_by("appt_datetime", "registered_subject__subject_identifier")
        )

    @property
    def stock_request(self):
        stock_request_id = self.kwargs.get("stock_request")
        try:
            stock_request = StockRequest.objects.get(id=stock_request_id)
        except ObjectDoesNotExist:
            stock_request = None
            messages.add_message(self.request, messages.ERROR, "Invalid stock request.")
        return stock_request

    @property
    def selected_assignment(self):
        assignment_id = self.kwargs.get("assignment")
        try:
            assignment = Assignment.objects.get(id=assignment_id)
        except ObjectDoesNotExist:
            assignment = None
        return assignment

    def get_next_subject_identifiers(self, count: int | None = None) -> list[str]:
        if self.selected_assignment:
            subject_identifiers = self.subject_identifiers
            if count:
                return [s for s in subject_identifiers[:count]]
            return [s for s in subject_identifiers]
        return []

    @property
    def stock_request_changelist_url(self) -> str:
        if self.stock_request:
            url = reverse("edc_pharmacy_admin:edc_pharmacy_stockrequest_changelist")
            return f"{url}?q={self.stock_request.request_identifier}"
        return "/"

    @staticmethod
    def get_assignment(assignment_id) -> Assignment | None:
        try:
            assignment = Assignment.objects.get(id=assignment_id)
        except ObjectDoesNotExist:
            assignment = None
        return assignment

    def redirect_on_has_duplicates(
        self,
        stock_codes: list[str],
        stock_request: StockRequest,
        assignment: Assignment,
    ) -> str | None:
        if len(stock_codes or []) != len(list(set(stock_codes or []))):
            messages.add_message(
                self.request,
                messages.ERROR,
                "Nothing saved. Duplicate codes detected in list. Please try again.",
            )
            return reverse(
                "edc_pharmacy:allocate_url",
                kwargs={
                    "stock_request": stock_request.id,
                    "assignment": assignment.id,
                },
            )
        return None

    def redirect_on_unconfirmed_stock_codes(
        self,
        stock_codes: list[str],
        stock_request: StockRequest,
        assignment: Assignment,
    ) -> str | None:
        if stock_codes:
            confirmed_codes = Stock.objects.filter(
                code__in=stock_codes,
                confirmation__isnull=False,
            ).values_list("code", flat=True)
            if len(confirmed_codes) != len(stock_codes):
                unconfirmed_codes = ", ".join(
                    [code for code in stock_codes if code not in confirmed_codes]
                )
                messages.add_message(
                    self.request,
                    messages.ERROR,
                    (
                        f"Nothing saved. Unconfirmed stock codes detected. "
                        f"Got {unconfirmed_codes}. "
                    ),
                )
                return reverse(
                    "edc_pharmacy:allocate_url",
                    kwargs={
                        "stock_request": stock_request.id,
                        "assignment": assignment.id,
                    },
                )
        return None

    def redirect_on_invalid_stock_codes(
        self,
        stock_codes: list[str],
        stock_request: StockRequest,
        assignment: Assignment,
    ) -> str | None:
        if stock_codes and Stock.objects.filter(code__in=stock_codes).count() != len(
            stock_codes
        ):
            valid_codes = Stock.objects.filter(code__in=stock_codes).values_list(
                "code", flat=True
            )
            invalid_codes = " ,".join(
                [code for code in stock_codes if code not in valid_codes]
            )
            messages.add_message(
                self.request,
                messages.ERROR,
                f"Nothing saved. Invalid codes detected. Got {invalid_codes}. ",
            )
            return reverse(
                "edc_pharmacy:allocate_url",
                kwargs={
                    "stock_request": stock_request.id,
                    "assignment": assignment.id,
                },
            )
        return None

    def redirect_on_has_multiple_container_types(
        self,
        stock_codes: list[str],
        stock_request: StockRequest,
        assignment: Assignment,
    ) -> str | None:
        if stock_codes and Stock.objects.filter(
            code__in=stock_codes, container=stock_request.container
        ).count() != len(stock_codes):
            messages.add_message(
                self.request,
                messages.ERROR,
                (
                    f"Nothing saved. Container mismatch for request. "
                    f"Expected `{stock_request.container}` "
                    f"only. See Stock request {stock_request.request_identifier} "
                ),
            )
            return reverse(
                "edc_pharmacy:allocate_url",
                kwargs={
                    "stock_request": stock_request.id,
                    "assignment": assignment.id,
                },
            )
        return None

    def redirect_on_stock_already_allocated(
        self,
        stock_codes: list[str],
        stock_request: StockRequest,
        assignment: Assignment,
    ) -> str | None:
        if (
            stock_codes
            and Stock.objects.filter(code__in=stock_codes, allocation__isnull=False).exists()
        ):
            allocated_stock_codes = []
            for stock in Stock.objects.filter(code__in=stock_codes):
                if stock.allocation:
                    allocated_stock_codes.append(stock.code)  # noqa: PERF401
            messages.add_message(
                self.request,
                messages.ERROR,
                f"Stock already allocated. Got {','.join(allocated_stock_codes)}.",
            )
            return reverse(
                "edc_pharmacy:allocate_url",
                kwargs={
                    "stock_request": stock_request.id,
                    "assignment": getattr(assignment, "id", None),
                },
            )
        return None

    def redirect_on_all_allocated_for_assignment(
        self, stock_request: StockRequest, assignment: Assignment
    ) -> str | None:
        if not stock_request.stockrequestitem_set.filter(
            allocation__isnull=True, assignment=assignment
        ).exists():
            messages.add_message(
                self.request,
                messages.INFO,
                _(
                    "All subjects in this stock request assigned '%(assignment)s' "
                    "medication have been allocated stock"
                )
                % {"assignment": assignment.display_name.upper()},
            )
            return reverse(
                "edc_pharmacy:allocate_url",
                kwargs={
                    "stock_request": stock_request.id,
                    "assignment": getattr(assignment, "id", None),
                },
            )
        return None

    def redirect_on_incorrect_stock_for_assignment(
        self,
        stock_codes: list[str],
        stock_request: StockRequest,
        assignment: Assignment,
    ) -> str | None:
        if stock_codes and Stock.objects.filter(
            code__in=stock_codes, lot__assignment=assignment
        ).count() != len(stock_codes):
            messages.add_message(
                self.request,
                messages.ERROR,
                (
                    "One or more stock codes are not for this assignment. "
                    f"Expected `{assignment.display_name}` only. Check your work."
                ),
            )
            return reverse(
                "edc_pharmacy:allocate_url",
                kwargs={
                    "stock_request": stock_request.id,
                    "assignment": getattr(assignment, "id", None),
                },
            )
        return None

    def get_counts(self, stock_request: StockRequest) -> tuple[int, int]:
        if self.selected_assignment:
            total_count = stock_request.stockrequestitem_set.all().count()
            groupby = (
                stock_request.stockrequestitem_set.values("assignment__name")
                .filter(allocation__isnull=True, allocation__stock__isnull=True)
                .annotate(count=Count("assignment__name"))
            )
            groupby = {dct["assignment__name"]: dct["count"] for dct in groupby}
            remaining_count = groupby.get(self.selected_assignment.name)
            return remaining_count or 0, total_count
        return 0, 0

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        stock_codes = request.POST.getlist("codes") if request.POST.get("codes") else None
        subject_identifiers = request.POST.get("subject_identifiers")
        assignment_id = request.POST.get("assignment")
        subject_identifiers = ast.literal_eval(subject_identifiers)
        stock_request = StockRequest.objects.get(id=kwargs.get("stock_request"))
        assignment = self.get_assignment(assignment_id)

        if url := self.redirect_on_all_allocated_for_assignment(stock_request, assignment):
            return HttpResponseRedirect(url)
        if url := self.redirect_on_has_duplicates(stock_codes, stock_request, assignment):
            return HttpResponseRedirect(url)
        if url := self.redirect_on_invalid_stock_codes(stock_codes, stock_request, assignment):
            return HttpResponseRedirect(url)
        if url := self.redirect_on_unconfirmed_stock_codes(
            stock_codes, stock_request, assignment
        ):
            return HttpResponseRedirect(url)
        if url := self.redirect_on_has_multiple_container_types(
            stock_codes, stock_request, assignment
        ):
            return HttpResponseRedirect(url)
        if url := self.redirect_on_incorrect_stock_for_assignment(
            stock_codes, stock_request, assignment
        ):
            return HttpResponseRedirect(url)

        if stock_codes and subject_identifiers and assignment:
            allocation_data = dict(zip(stock_codes, subject_identifiers, strict=False))
            try:
                allocated, not_allocated = allocate_stock(
                    stock_request,
                    allocation_data,
                    allocated_by=request.user.username,
                    user_created=request.user.username,
                    created=timezone.now(),
                )
            except AllocationError as e:
                messages.add_message(request, messages.ERROR, str(e))
            except InsufficientStockError as e:
                messages.add_message(request, messages.ERROR, str(e))
            else:
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    f"Allocated {len(allocated)} stock records. "
                    f"Skipped {len(not_allocated)}. Skipped {', '.join(not_allocated)}",
                )
            if self.get_next_subject_identifiers():
                url = reverse(
                    "edc_pharmacy:allocate_url",
                    kwargs={
                        "stock_request": stock_request.id,
                        "assignment": assignment.id,
                    },
                )
                return HttpResponseRedirect(url)
            return HttpResponseRedirect(self.stock_request_changelist_url)
        url = reverse(
            "edc_pharmacy:allocate_url",
            kwargs={
                "stock_request": stock_request.id,
                "assignment": getattr(assignment, "id", None),
            },
        )
        return HttpResponseRedirect(url)



# Generated by Django 5.2.10 on 2026-01-25 01:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0121_confirmation_code_historicalconfirmation_code_and_more"),
    ]

    operations = []



from django.db import models
from edc_model_fields.fields import OtherCharField

from ..choices import APPT_STATUS, APPT_TIMING, DEFAULT_APPT_REASON_CHOICES
from ..constants import NEW_APPT, ONTIME_APPT


class AppointmentFieldsModelMixin(models.Model):
    timepoint = models.DecimalField(
        null=True, decimal_places=1, max_digits=6, help_text="timepoint from schedule"
    )

    timepoint_datetime = models.DateTimeField(
        null=True, help_text="Unadjusted datetime calculated from visit schedule"
    )

    appt_close_datetime = models.DateTimeField(
        null=True,
        help_text=(
            "timepoint_datetime adjusted according to the nearest "
            "available datetime for this facility"
        ),
    )

    facility_name = models.CharField(
        max_length=25,
        help_text="set by model that creates appointments, e.g. Enrollment",
    )

    appt_datetime = models.DateTimeField(
        verbose_name="Appointment date and time", db_index=True
    )

    appt_type = models.ForeignKey(
        "edc_appointment.AppointmentType",
        verbose_name="Appointment type",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        help_text="",
    )

    appt_type_other = OtherCharField(
        verbose_name="If other appointment type, please specify ...",
    )

    appt_status = models.CharField(
        verbose_name="Status",
        choices=APPT_STATUS,
        max_length=25,
        default=NEW_APPT,
        db_index=True,
        help_text=(
            "If the visit has already begun, only 'in progress', "
            "'incomplete' or 'done' are valid options. Only unscheduled appointments "
            "may be cancelled."
        ),
    )

    appt_reason = models.CharField(
        verbose_name="Reason for appointment",
        max_length=25,
        choices=DEFAULT_APPT_REASON_CHOICES,
        help_text=(
            "The reason for visit from the visit report will be validated against "
            "this response. Refer to the protocol documentation for the definition "
            "of a scheduled appointment."
        ),
    )

    appt_timing = models.CharField(
        verbose_name="Timing",
        max_length=25,
        choices=APPT_TIMING,
        default=ONTIME_APPT,
        help_text=(
            "If late, you may also be required to complete a protocol incident report. "
            "Extended window period may be allowed for the final appointment. "
            "Refer to the protocol documentation for the allowed window periods "
            "of scheduled appointments."
        ),
    )

    comment = models.CharField("Comment", max_length=250, blank=True, default="")

    is_confirmed = models.BooleanField(default=False, editable=False)

    ignore_window_period = models.BooleanField(default=False, editable=False)

    class Meta:
        abstract = True



import usb
import usb.core
import usb.util
from django.http import JsonResponse


class FindDevices:
    def __init__(self, class_):
        self._class = class_

    def __call__(self, device):
        if device.bDeviceClass == self._class:
            return True
        for cfg in device:
            intf = usb.util.find_descriptor(cfg, bInterfaceClass=self._class)
            if intf is not None:
                return True
        return False


def list_usb_printers():
    printers = usb.core.find(find_all=1, custom_match=FindDevices(7))
    return [f"{dev.manufacturer} {dev.product}" for dev in printers]


def print_label_usb(request):
    """
    see pyusb
    https://github.com/pyusb/pyusb/blob/master/docs/tutorial.rst

    usb module requires ...
        linux: sudo apt-get install libusb-1.0-0-dev
        mac: brew install libusb
        windows: download the libusb DLL from the libusb
            website and place it in your system directory
    """
    if request.method == "POST":
        zpl_code = request.body.decode("utf-8")
        dev = usb.core.find(idVendor=0x0A5F)  # Replace with your printer's vendor ID
        if dev is None:
            return JsonResponse({"error": "Printer not found"}, status=404)

        dev.set_configuration()
        cfg = dev.get_active_configuration()
        intf = cfg[(0, 0)]
        ep = usb.util.find_descriptor(
            intf,
            custom_match=lambda e: usb.util.endpoint_direction(e.bEndpointAddress)
            == usb.util.ENDPOINT_OUT,
        )

        if ep is None:
            return JsonResponse({"error": "Endpoint not found"}, status=404)

        ep.write(zpl_code)
        return JsonResponse({"message": "Label sent to printer"}, status=200)
    return JsonResponse({"error": "Invalid request method"}, status=400)



# Generated by Django 5.1 on 2024-08-15 22:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0015_alter_note_status"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="note",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                    "viewallsites",
                ),
                "verbose_name": "Note",
                "verbose_name_plural": "Notes",
            },
        ),
        migrations.AlterModelOptions(
            name="qareportlog",
            options={
                "default_permissions": ("view", "export", "viewallsites"),
                "verbose_name": "QA Report Log",
                "verbose_name_plural": "QA Report Logs",
            },
        ),
    ]



from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..models import Container


def format_qty(qty: Decimal, container: Container):
    qty = 0 if qty is None else qty
    if container.unit_qty_places == 0:
        return str(int(qty))
    if container.unit_qty_places == 1:
        return f"{qty:0.1f}"
    return f"{qty:0.2f}"


__all__ = ["format_qty"]



from __future__ import annotations

import calendar
import sys
import warnings
from collections.abc import Callable
from datetime import datetime
from typing import TYPE_CHECKING, Any

from clinicedc_constants import CLINIC, NOT_APPLICABLE, OK
from clinicedc_constants import ERROR as ERROR_CODE
from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.admin.utils import NotRelationField, get_model_from_relation
from django.contrib.messages import ERROR, SUCCESS
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    ValidationError,
)
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.db.models import Count, ProtectedError
from django.urls import reverse
from django.utils.translation import gettext as _

from edc_dashboard.url_names import url_names
from edc_form_validators import INVALID_ERROR
from edc_metadata.constants import CRF, REQUIRED, REQUISITION
from edc_metadata.utils import (
    get_crf_metadata_model_cls,
    get_requisition_metadata_model_cls,
    has_keyed_metadata,
)
from edc_utils.date import to_local, to_utc
from edc_utils.text import convert_php_dateformat
from edc_visit_schedule.exceptions import (
    ScheduledVisitWindowError,
    UnScheduledVisitWindowError,
)
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.utils import get_default_max_visit_window_gap, is_baseline
from edc_visit_tracking.utils import get_allow_missed_unscheduled_appts

from .choices import DEFAULT_APPT_REASON_CHOICES
from .constants import (
    CANCELLED_APPT,
    COMPLETE_APPT,
    EXTENDED_APPT,
    INCOMPLETE_APPT,
    MISSED_APPT,
    NEW_APPT,
    SCHEDULED_APPT,
    SKIPPED_APPT,
    UNSCHEDULED_APPT,
)
from .exceptions import (
    AppointmentBaselineError,
    AppointmentDatetimeError,
    AppointmentMissingValuesError,
    AppointmentWindowError,
    UnscheduledAppointmentError,
)

if TYPE_CHECKING:
    from decimal import Decimal

    from django.db.models import QuerySet

    from edc_crf.model_mixins import CrfModelMixin as Base
    from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin

    from .models import Appointment, AppointmentType

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        appointment: Appointment


class AppointmentDateWindowPeriodGapError(Exception):
    pass


class AppointmentAlreadyStarted(Exception):  # noqa: N818
    pass


def get_appointment_model_name() -> str:
    return "edc_appointment.appointment"


def get_appointment_model_cls() -> type[Appointment]:
    return django_apps.get_model(get_appointment_model_name())


def get_appointment_type_model_name() -> str:
    return "edc_appointment.appointmenttype"


def get_appointment_type_model_cls() -> type[AppointmentType]:
    return django_apps.get_model(get_appointment_type_model_name())


def get_allow_skipped_appt_using() -> dict[str, tuple[str, str]]:
    """Return dict from settings or empty dictionary.

    For example:
         = {"my_app.crfone": ("next_appt_date", "next_visit_code")}
    """
    dct = getattr(settings, "EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING", {})
    if len(dct.keys()) > 1:
        raise ImproperlyConfigured(
            "Only one model may be specified. See "
            f"settings.EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING. Got {dct}."
        )
    return dct


def get_allow_unscheduled_for_appt_using() -> dict[str, tuple[str, str]]:
    """Return dict from settings or empty dictionary.

    For example:
         = {"my_app.crfone": ("next_appt_date", "next_visit_code")}
    """
    dct = getattr(settings, "EDC_APPOINTMENT_ALLOW_UNSCHEDULED_FOR_APPT_USING", {})
    if len(dct.keys()) > 1:
        raise ImproperlyConfigured(
            "Only one model may be specified. See "
            f"settings.EDC_APPOINTMENT_ALLOW_UNSCHEDULED_FOR_APPT_USING. Got {dct}."
        )
    return dct


def get_max_months_to_next_appointment():
    return getattr(settings, "EDC_APPOINTMENT_MAX_MONTHS_TO_NEXT_APPT", 6)


def allow_clinic_on_weekend():
    return getattr(settings, "EDC_APPOINTMENT_ALLOW_CLINIC_ON_WEEKENDS", False)


def get_max_months_to_next_appointment_as_rdelta():
    max_months = get_max_months_to_next_appointment()
    return relativedelta(months=max_months)


def raise_on_appt_may_not_be_missed(
    appointment: Appointment = None,
    appt_timing: str | None = None,
):
    if appointment.id:
        appt_timing = appt_timing or appointment.appt_timing
        if (
            appt_timing
            and appt_timing in [MISSED_APPT, NOT_APPLICABLE]
            and is_baseline(instance=appointment)
        ):
            errmsg = "Invalid. A baseline appointment may not be reported as missed"
            if appt_timing == NOT_APPLICABLE:
                errmsg = "This field is applicable"
            raise AppointmentBaselineError(errmsg)
        if (
            appointment.visit_code_sequence is not None
            and appt_timing
            and appointment.visit_code_sequence > 0
            and appt_timing == MISSED_APPT
            and not get_allow_missed_unscheduled_appts()
        ):
            raise UnscheduledAppointmentError(
                "Invalid. An unscheduled appointment may not be reported as missed."
                "Try to cancel the appointment instead. "
            )


def get_appointment_form_meta_options() -> dict:
    return getattr(
        settings, "EDC_APPOINTMENT_FORM_META_OPTIONS", dict(labels={}, help_texts={})
    )


def get_appt_reason_choices() -> tuple[str, ...]:
    """Returns a customized tuple of choices otherwise the default.

    Note: You can only change the right side of any tuple.

    For example:
      ((SCHEDULED_APPT, "some custom text"),
        (UNSCHEDULED_APPT, "some custom text"))

    See also: formfield_for_choice_field in modeladmin class.
    """

    settings_attr = "EDC_APPOINTMENT_APPT_REASON_CHOICES"
    appt_reason_choices = getattr(settings, settings_attr, DEFAULT_APPT_REASON_CHOICES)
    keys = sorted([choice[0] for choice in appt_reason_choices])
    if SCHEDULED_APPT not in keys or UNSCHEDULED_APPT not in keys:
        raise ImproperlyConfigured(
            "Invalid value for EDC_APPOINTMENT_APPT_REASON_CHOICES. "
            f"Expected a choices tuple with keys `{SCHEDULED_APPT}` and `{UNSCHEDULED_APPT}`. "
            f"See {settings_attr}."
        )
    return appt_reason_choices


def get_appt_type_default() -> str:
    """Returns the default appointment type name."""
    return getattr(settings, "EDC_APPOINTMENT_APPT_TYPE_DEFAULT", CLINIC)


def get_appt_reason_default() -> str:
    """Returns the default appointment reason."""
    value = getattr(settings, "EDC_APPOINTMENT_APPT_REASON_DEFAULT", None)
    if not value:
        value = getattr(settings, "EDC_APPOINTMENT_DEFAULT_APPT_REASON", None)
        warnings.warn(
            "Settings attribute `EDC_APPOINTMENT_DEFAULT_APPT_REASON` has "
            "been deprecated in favor of `EDC_APPOINTMENT_APPT_REASON_DEFAULT`. ",
            DeprecationWarning,
            stacklevel=2,
        )
    return value or SCHEDULED_APPT


def cancelled_appointment(appointment: Appointment) -> None:
    """Try to delete subject visit if appt status = CANCELLED_APPT"""
    try:
        cancelled = appointment.appt_status == CANCELLED_APPT
    except AttributeError as e:
        if "appt_status" not in str(e):
            raise
    else:
        if (
            cancelled
            and appointment.visit_code_sequence > 0
            and "historical" not in appointment._meta.label_lower
            and not appointment.crf_metadata_keyed_exists
            and not appointment.requisition_metadata_keyed_exists
        ):
            try:
                related_visit = appointment.related_visit_model_cls().objects.get(
                    appointment=appointment
                )
            except ObjectDoesNotExist:
                appointment.delete()
            else:
                with transaction.atomic():
                    try:
                        related_visit.delete()
                    except ProtectedError:
                        pass
                    else:
                        appointment.delete()


def missed_appointment(appointment: Appointment) -> None:
    """Try to create_missed_visit_from_appointment if
    appt_status == missed.
    """
    try:
        missed = appointment.appt_timing == MISSED_APPT
    except AttributeError as e:
        if "appt_timing" not in str(e):
            raise
    else:
        if (
            missed
            and appointment.visit_code_sequence == 0
            and "historical" not in appointment._meta.label_lower
        ):
            try:
                appointment.create_missed_visit_from_appointment()
            except AttributeError as e:
                if "create_missed_visit" not in str(e):
                    raise


def reset_visit_code_sequence_or_pass(
    subject_identifier: str,
    visit_schedule_name: str,
    schedule_name: str,
    visit_code: str,
    appointment: Appointment | None = None,
    write_stdout: bool | None = None,
) -> Appointment | None:
    """Validate the order of the appointment visit code sequences
    relative to the appt_datetime and reset the visit code sequences
    if needed.

    Delete and recreate metadata

    Also do the same for the `related_visit`, if it exists.
    """
    opts = dict(
        subject_identifier=subject_identifier,
        visit_schedule_name=visit_schedule_name,
        schedule_name=schedule_name,
        visit_code=visit_code,
    )
    qs = get_appointment_model_cls().objects.filter(**opts).order_by("appt_datetime")
    expected = list(range(0, qs.count()))
    actual = [o.visit_code_sequence for o in qs]
    if actual != expected:
        if write_stdout:
            sys.stdout.write(
                "     - Resetting for "
                f"{subject_identifier} {visit_code}: {actual=} {expected=} ...\n"
            )
        # reset visit code sequence for this visit code
        get_crf_metadata_model_cls().objects.filter(visit_code_sequence__gt=0, **opts).delete()
        get_requisition_metadata_model_cls().objects.filter(
            visit_code_sequence__gt=0, **opts
        ).delete()

        with transaction.atomic():
            # set appt and related visit visit_code_sequences to the
            # negative of the current value
            for obj in get_appointment_model_cls().objects.filter(
                visit_code_sequence__gt=0, **opts
            ):
                obj.visit_code_sequence = obj.visit_code_sequence * -1
                obj.save_base(update_fields=["visit_code_sequence"])
                if getattr(obj, "related_visit", None):
                    obj.related_visit.visit_code_sequence = obj.visit_code_sequence
                    obj.related_visit.save_base(update_fields=["visit_code_sequence"])
                    obj.related_visit.metadata_create()

            # reset sequence order by appt_datetime
            for index, obj in enumerate(
                get_appointment_model_cls()
                .objects.filter(visit_code_sequence__lt=0, **opts)
                .order_by("appt_datetime"),
                start=1,
            ):
                obj.visit_code_sequence = index
                obj.save_base(update_fields=["visit_code_sequence"])
                if getattr(obj, "related_visit", None):
                    obj.related_visit.visit_code_sequence = index
                    obj.related_visit.save_base(update_fields=["visit_code_sequence"])
                    obj.related_visit.metadata_create()

        if appointment:
            # refresh the given appt if not None since
            # appointment visit_code_sequence may have changed
            appointment = get_appointment_model_cls().objects.get(id=appointment.id)
    return appointment


def reset_visit_code_sequence_for_subject(
    subject_identifier: str,
    visit_schedule_name: str,
    schedule_name: str,
) -> None:
    """Resets / validates appointment `visit code sequences` for any
    `visit code` with unscheduled appointments for the given subject
     and schedule.

     Wrapper for function `reset_visit_code_sequence_or_pass`.
    """
    ann = (
        get_appointment_model_cls()
        .objects.values("visit_code")
        .filter(subject_identifier=subject_identifier, visit_code_sequence__gt=0)
        .annotate(Count("visit_code"))
    )
    for visit_code in [obj.get("visit_code") for obj in ann]:
        reset_visit_code_sequence_or_pass(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code=visit_code,
        )


def delete_appointment_in_sequence(appointment: Any, from_post_delete=None) -> None:
    if not from_post_delete:
        with transaction.atomic():
            appointment.delete()
        reset_visit_code_sequence_or_pass(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
        )


def update_appt_status(appointment: Appointment, save: bool | None = None):
    """Sets appt_status, and if save is True, calls save_base().

    This is useful if checking `appt_status` is correct
    relative to the visit tracking model and CRFs and
    requisitions
    """
    if appointment.appt_status in (CANCELLED_APPT, SKIPPED_APPT):
        pass
    elif not appointment.related_visit:
        appointment.appt_status = NEW_APPT
    elif (
        appointment.crf_metadata_required_exists
        or appointment.requisition_metadata_required_exists
    ):
        appointment.appt_status = INCOMPLETE_APPT
    else:
        appointment.appt_status = COMPLETE_APPT
    if save:
        appointment.save_base(update_fields=["appt_status"])
        appointment.refresh_from_db()
    return appointment


def get_previous_appointment(
    appointment: Appointment, include_interim: bool | None = None
) -> Appointment | None:
    """Returns the previous appointment model instance,
    or None, in this schedule.

    Keywords:
        * include_interim: include interim appointments
          (e.g. those where visit_code_sequence != 0)

    See also: `AppointmentMethodsModelMixin`
    """
    check_appointment_required_values_or_raise(appointment)
    opts: dict[str, str | int | Decimal] = dict(
        subject_identifier=appointment.subject_identifier,
        visit_schedule_name=appointment.visit_schedule_name,
        schedule_name=appointment.schedule_name,
    )
    if include_interim:
        if appointment.visit_code_sequence != 0:
            opts.update(
                visit_code_sequence__lt=appointment.visit_code_sequence,
                timepoint__lte=appointment.timepoint,
            )
        else:
            opts.update(timepoint__lt=appointment.timepoint)
    elif not include_interim:
        opts.update(
            visit_code_sequence=0,
            timepoint__lt=appointment.timepoint,
        )

    appointments: QuerySet[Appointment] = (
        appointment.__class__.objects.filter(**opts)
        .exclude(id=appointment.id)
        .order_by("timepoint", "visit_code_sequence")
    )

    try:
        previous_appt = appointments.reverse()[0]
    except IndexError:
        previous_appt = None
    return previous_appt


def get_next_appointment(appointment: Appointment, include_interim=None) -> Appointment | None:
    """Returns the next appointment model instance,
    or None, in this schedule.

    Keywords:
        * include_interim: include interim appointments
          (e.g. those where visit_code_sequence != 0)

    See also: `AppointmentMethodsModelMixin`
    """
    next_appt: Appointment | None = None
    check_appointment_required_values_or_raise(appointment)
    opts: dict[str, str | int | Decimal] = dict(
        subject_identifier=appointment.subject_identifier,
        visit_schedule_name=appointment.visit_schedule_name,
        schedule_name=appointment.schedule_name,
    )
    if include_interim:
        break_on_next = False
        for obj in appointment.__class__.objects.filter(
            timepoint__gte=appointment.timepoint, **opts
        ).order_by("timepoint", "visit_code_sequence"):
            if break_on_next:
                next_appt = obj
                break
            if obj.id == appointment.id:
                break_on_next = True
    elif not include_interim:
        opts.update(
            timepoint__gt=appointment.timepoint,
            visit_code_sequence=0,
        )
        next_appt = (
            appointment.__class__.objects.filter(**opts)
            .exclude(id=appointment.id)
            .order_by("timepoint", "visit_code_sequence")
        ).first()
    return next_appt


def check_appointment_required_values_or_raise(appointment: Appointment) -> None:
    if (
        not appointment.visit_schedule_name
        or not appointment.schedule_name
        or not appointment.visit_code
        or appointment.visit_code_sequence is None
        or appointment.timepoint is None
    ):
        raise AppointmentMissingValuesError(
            f"Appointment instance is missing required values. See {appointment}."
        )


def raise_on_appt_datetime_not_in_window(
    appointment: Appointment,
    appt_datetime: datetime | None = None,
    baseline_timepoint_datetime: datetime | None = None,
) -> None:
    if appointment.appt_status != CANCELLED_APPT and not is_baseline(instance=appointment):
        baseline_timepoint_datetime = baseline_timepoint_datetime or (
            appointment.__class__.objects.first_appointment(
                subject_identifier=appointment.subject_identifier,
                visit_schedule_name=appointment.visit_schedule_name,
                schedule_name=appointment.schedule_name,
            ).timepoint_datetime
        )
        try:
            appointment.schedule.datetime_in_window(
                dt=appt_datetime or appointment.appt_datetime,
                baseline_timepoint_datetime=baseline_timepoint_datetime,
                timepoint_datetime=appointment.timepoint_datetime,
                visit=appointment.visit,
                next_visit=getattr(appointment.next, "visit", None),
                visit_code_sequence=appointment.visit_code_sequence,
            )
        except ScheduledVisitWindowError as e:
            msg = str(e)
            msg = msg.replace("Invalid datetime", "Invalid appointment datetime (S)")
            raise AppointmentWindowError(msg) from e
        except UnScheduledVisitWindowError as e:
            msg = str(e)
            msg = msg.replace("Invalid datetime", "Invalid appointment datetime (U)")
            raise AppointmentWindowError(msg) from e


def get_window_gap_days(appointment) -> int:
    """Return the number of days betwen this visit's upper and the
    next visit's lower.

    See get_default_max_visit_window_gap and settings attr.
    """
    if not appointment.next:
        gap_days = 0
    else:
        gap_days = abs(
            (appointment.timepoint_datetime + appointment.visit.rupper)
            - (appointment.next.timepoint_datetime - appointment.next.visit.rlower)
        ).days
    return gap_days


def appt_datetime_in_gap(appointment: Appointment, suggested_appt_datetime: datetime) -> bool:
    """Return True if datetime falls in a gap between this and the
    next appointment window.
    """
    in_gap = False
    if get_window_gap_days(appointment) > 0:
        next_lower_datetime = (
            appointment.next.timepoint_datetime - appointment.next.visit.rlower
        )
        upper_datetime = appointment.timepoint_datetime + appointment.visit.rupper
        if upper_datetime < suggested_appt_datetime < next_lower_datetime:
            in_gap = True
    return in_gap


def get_max_window_gap_to_lower(appointment) -> int:
    if (
        appointment.visit.max_window_gap_to_lower is not None
        and appointment.visit.max_window_gap_to_lower < get_default_max_visit_window_gap()
    ):
        max_gap = appointment.visit.max_window_gap_to_lower
    else:
        max_gap = get_default_max_visit_window_gap()
    return max_gap


def appt_datetime_in_next_window_adjusted_for_gap(
    appointment: Appointment, suggested_appt_datetime: datetime
) -> bool:
    """Returns True if `suggest_datetime` falls between the
    NEXT appointment's lower and upper window period datetime after
    adding gap_days to the lower datetime.
    """
    in_window = False
    gap_days = get_window_gap_days(appointment)
    max_gap = get_max_window_gap_to_lower(appointment)
    gap_days = min(gap_days, max_gap)
    if gap_days > 0:
        next_lower_datetime = (
            appointment.next.timepoint_datetime
            - appointment.next.visit.rlower
            - relativedelta(days=gap_days)
        )
        next_upper_datetime = (
            appointment.next.timepoint_datetime + appointment.next.visit.rupper
        )
        if next_lower_datetime <= suggested_appt_datetime <= next_upper_datetime:
            in_window = True
    return in_window


def get_appointment_by_datetime(
    suggested_appt_datetime: datetime,
    subject_identifier: str,
    visit_schedule_name: str,
    schedule_name: str,
    raise_if_in_gap: bool | None = None,
) -> Appointment | None:
    """Returns an appointment where the suggested datetime falls
    within the window period.

    * Returns None if no appointment is found.
    * Raises an exception if there is a gap between upper and lower
      boundaries and the date falls within the gap.
    """
    appointment = None
    raise_if_in_gap = True if raise_if_in_gap is None else raise_if_in_gap
    appointments = (
        django_apps.get_model("edc_appointment.appointment")
        .objects.filter(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code_sequence=0,
        )
        .order_by("timepoint_datetime")
    )
    for appointment in appointments:
        if appointment.appt_status == CANCELLED_APPT or is_baseline(appointment):
            continue
        try:
            raise_on_appt_datetime_not_in_window(
                appointment, appt_datetime=suggested_appt_datetime
            )
        except AppointmentWindowError as e:
            in_gap = appt_datetime_in_gap(appointment, suggested_appt_datetime)
            in_next_window_adjusted = appt_datetime_in_next_window_adjusted_for_gap(
                appointment, suggested_appt_datetime
            )
            if in_gap and raise_if_in_gap:
                dt = suggested_appt_datetime.strftime(
                    convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                )
                raise AppointmentDateWindowPeriodGapError(
                    f"Date falls in a `window period gap` between {appointment.visit_code} "
                    f"and {appointment.next.visit_code}. Got {dt}."
                ) from e
            if (
                in_gap
                and in_next_window_adjusted
                and appointment.next.visit.add_window_gap_to_lower
            ):
                appointment = appointment.next  # noqa: PLW2901
                break
            if (
                in_gap
                and not in_next_window_adjusted
                and appointment.next.visit.add_window_gap_to_lower
            ):
                appointment = None  # noqa: PLW2901
                break
            appointment = appointment.next  # noqa: PLW2901
        else:
            break
    return appointment


def reset_appointment(appointment: Appointment, **kwargs):
    """Reset appointment but only if appointment has not started.

    Will overwrite the default field values with values
    from kwargs.

    If field in kwargs refers to a field class with a related_model
    and the value is not found, the field will be set to None without
    raising an ObjectDoesNotExist exception.
    """
    if has_keyed_metadata(appointment) or appointment.related_visit:
        raise AppointmentAlreadyStarted(
            f"Unable to reset. Appointment already started. Got {appointment}."
        )
    defaults = dict(
        appt_status=appointment._meta.get_field("appt_status").default,
        appt_timing=appointment._meta.get_field("appt_timing").default,
        appt_type=None,
        appt_type_other="",
        appt_datetime=appointment.timepoint_datetime,
        comment="",
    )
    defaults.update(**kwargs)
    for k, v in defaults.items():
        try:
            related_model = get_model_from_relation(appointment._meta.get_field(k))
        except NotRelationField:
            setattr(appointment, k, v)
        else:
            try:
                setattr(appointment, k, related_model.objects.get(name=v))
            except ObjectDoesNotExist:
                setattr(appointment, k, None)
    appointment.save_base(update_fields=[*defaults.keys()])


def skip_appointment(appointment: Appointment, comment: str | None = None):
    """Set appointment to `SKIPPED_APPT` if appointment has not
    started.
    """
    if has_keyed_metadata(appointment) or appointment.related_visit:
        raise AppointmentAlreadyStarted(
            f"Unable to skip. Appointment already started. Got {appointment}."
        )
    reset_appointment(
        appointment,
        appt_status=SKIPPED_APPT,
        appt_timing=NOT_APPLICABLE,
        appt_type=NOT_APPLICABLE,
        comment=comment or "",
    )


def get_unscheduled_appointment_url(appointment: Appointment = None) -> str:
    """Returns a url for the unscheduled appointment."""
    dashboard_url = url_names.get("subject_dashboard_url")
    unscheduled_appointment_url = "edc_appointment:unscheduled_appointment_url"
    kwargs = dict(
        subject_identifier=appointment.subject_identifier,
        visit_schedule_name=appointment.visit_schedule_name,
        schedule_name=appointment.schedule_name,
        visit_code=appointment.visit_code,
        visit_code_sequence=appointment.visit_code_sequence + 1,
        timepoint=appointment.timepoint,
    )
    kwargs.update(visit_code_sequence=str(appointment.visit_code_sequence + 1))
    kwargs.update(redirect_url=dashboard_url)
    return reverse(unscheduled_appointment_url, kwargs=kwargs)


def update_appt_status_for_timepoint(related_visit: RelatedVisitModel) -> None:
    """Only check COMPLETE_APPT and INCOMPLETE_APPT against metadata."""
    if related_visit.appointment.appt_status == COMPLETE_APPT:
        if (
            related_visit.metadata[CRF].filter(entry_status=REQUIRED).exists()
            or related_visit.metadata[REQUISITION].filter(entry_status=REQUIRED).exists()
        ):
            related_visit.appointment.appt_status = INCOMPLETE_APPT
            related_visit.appointment.save_base(update_fields=["appt_status"])
    elif related_visit.appointment.appt_status == INCOMPLETE_APPT and (
        not related_visit.metadata[CRF].filter(entry_status=REQUIRED).exists()
        and not related_visit.metadata[REQUISITION].filter(entry_status=REQUIRED).exists()
    ):
        related_visit.appointment.appt_status = COMPLETE_APPT
        related_visit.appointment.save_base(update_fields=["appt_status"])


def offschedule(
    subject_identifier: str,
    offschedule_model: str,
    request: WSGIRequest,
    verbose: bool | None = None,
):
    try:
        django_apps.get_model(offschedule_model).objects.get(
            subject_identifier=subject_identifier
        )
    except ObjectDoesNotExist:
        retval = False
    else:
        retval = True
        msg = _(
            "Unable to refreshing appointments. Subject '%(subject_identifier)s' "
            "is off schedule."
        ) % dict(subject_identifier=subject_identifier)
        if request:
            messages.add_message(
                request=request,
                level=ERROR,
                message=msg,
            )
        elif verbose:
            sys.stdout.write(f"{msg}\n")
    return retval


def refresh_appointments(
    subject_identifier: str,
    visit_schedule_name: str,
    schedule_name: str,
    request: WSGIRequest | None = None,
    warn_only: bool | None = None,
    skip_get_current_site: bool | None = None,
) -> tuple[str, str]:
    status = OK
    visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name)
    schedule = visit_schedule.schedules.get(schedule_name)
    if not offschedule(subject_identifier, schedule.offschedule_model, request):
        try:
            schedule.refresh_schedule(
                subject_identifier,
                skip_get_current_site=skip_get_current_site,
            )
        except AppointmentDatetimeError as e:
            if request and not warn_only:
                status = ERROR_CODE
                messages.add_message(
                    request=request,
                    level=ERROR,
                    message=_(
                        "An error was encountered when refreshing appointments. "
                        "Contact your administrator. Got '%(error_msg)s'."
                    )
                    % dict(error_msg=str(e)),
                )
            elif warn_only:
                warnings.warn(str(e), stacklevel=2)
            else:
                raise
        else:
            for appointment in get_appointment_model_cls().objects.filter(
                subject_identifier=subject_identifier,
                visit_schedule_name=visit_schedule_name,
                schedule_name=schedule_name,
            ):
                if appointment.related_visit:
                    update_appt_status_for_timepoint(appointment.related_visit)
        if status == OK and request:
            messages.add_message(
                request,
                SUCCESS,
                _("The appointments for %(subject_identifier)s have been refreshed")
                % dict(subject_identifier=subject_identifier),
            )
    return subject_identifier, status


def validate_date_is_on_clinic_day(
    cleaned_data: dict, clinic_days: list[int], raise_validation_error: Callable | None = None
):
    raise_validation_error = raise_validation_error or ValidationError
    if cleaned_data.get("appt_date"):
        try:
            appt_date = to_local(cleaned_data.get("appt_date")).date()
        except AttributeError:
            appt_date = cleaned_data.get("appt_date")
        try:
            report_date = to_local(cleaned_data.get("report_datetime")).date()
        except AttributeError:
            report_date = cleaned_data.get("report_datetime")
        day_abbr = calendar.weekheader(3).split(" ")
        if appt_date == report_date:
            raise raise_validation_error(
                {"appt_date": "Cannot be equal to the report datetime"}, INVALID_ERROR
            )
        if appt_date <= report_date:
            raise raise_validation_error(
                {"appt_date": "Cannot be before the report datetime"}, INVALID_ERROR
            )
        if not allow_clinic_on_weekend() and calendar.weekday(
            appt_date.year, appt_date.month, appt_date.day
        ) in [calendar.SATURDAY, calendar.SUNDAY]:
            raise raise_validation_error(
                {
                    "appt_date": _("Expected %(mon)s-%(fri)s. Got %(day)s")
                    % {
                        "mon": day_abbr[calendar.MONDAY],
                        "fri": day_abbr[calendar.FRIDAY],
                        "day": day_abbr[
                            calendar.weekday(appt_date.year, appt_date.month, appt_date.day)
                        ],
                    }
                },
                INVALID_ERROR,
            )
        if (
            clinic_days
            and calendar.weekday(appt_date.year, appt_date.month, appt_date.day)
            not in clinic_days
        ):
            days_str = [day_abbr[d] for d in clinic_days]
            raise raise_validation_error(
                {
                    "appt_date": _(
                        "Invalid clinic day. Expected %(expected)s. Got %(day_abbr)s"
                    )
                    % {
                        "expected": ", ".join(days_str),
                        "day_abbr": appt_date.strftime("%A"),
                    }
                },
                INVALID_ERROR,
            )


def allow_extended_window_period(
    proposed_appt_timing: str,
    proposed_appt_datetime: datetime,
    appointment: Appointment,
):
    """Allow to save an appt with an appt datetime beyond the upper
    boundary of the window period.

    The visit object (not the model instance) must have
    rupper_extended set to a relativedelta(). Also, the appt_timing
    must be EXTENDED_APPT. There may NOT be a 'next' appointment to
    avoid clashes with the next appointment lower boundary.

    The scenario is rare, but it may be that some additional data
    comes in months after the expected upper window period of the
    last visit. Such data may be allowed if the PI approves.

    In the case of the 2025 META Phase III trial, some subjects
    extended the 36m schedule to 48m, others did not. For both
    groups, supplies for the OGTT were unavailable at their final
    visit. The PI agreed to allow patients to come in for the OGTT
    months beyond the upper bounary of the final appt's defined
    window period (36m or 48m).
    """
    return bool(
        proposed_appt_timing == EXTENDED_APPT
        and not appointment.next
        and appointment.visit.rupper_extended
        and appointment.timepoint_datetime
        <= to_utc(proposed_appt_datetime)
        <= appointment.timepoint_datetime + appointment.visit.rupper_extended
    )



from edc_navbar import Navbar, NavbarItem, site_navbars

navbar = Navbar(name="edc_review_dashboard")

navbar_item = NavbarItem(
    name="subject_review",
    label="Review",
    title="Subject Review",
    codename="edc_review_dashboard.view_subject_review_listboard",
    url_names_key="subject_review_listboard_url",
)

navbar.register(navbar_item)

site_navbars.register(navbar)



# Generated by Django 4.2.3 on 2023-07-20 03:10

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_locator", "0032_alter_subjectlocator_managers"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="subjectlocator",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Subject Locator",
                "verbose_name_plural": "Subject Locators",
            },
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from .form_runner import FormRunner

if TYPE_CHECKING:
    import uuid

__all__ = ["BaseFormRunnerBySrcId", "FormRunnerBySrcId"]


class BaseFormRunnerBySrcId:
    def __init__(
        self, src_id: uuid.UUID, model_name: str | None, verbose: bool | None = None
    ) -> None:
        self.src_id = src_id
        super().__init__(
            src_filter_options=dict(id=self.src_id),
            model_name=model_name,
            verbose=verbose,
        )

    def run_all(self):
        raise NotImplementedError

    def run_one(self) -> None:
        src_obj = self.src_model_cls.objects.get(id=self.src_id)
        super().run_one(src_obj)


class FormRunnerBySrcId(BaseFormRunnerBySrcId, FormRunner):
    def __init__(self, src_id: uuid.UUID, model_name: str | None, verbose: bool | None = None):
        super().__init__(src_id=src_id, model_name=model_name, verbose=verbose)



# Generated by Django 5.2.10 on 2026-01-25 05:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0128_auto_20260125_0741"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalrepackrequest",
            name="override_container_unit_qty",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="repackrequest",
            name="override_container_unit_qty",
            field=models.BooleanField(default=False),
        ),
    ]



class PrescriptionApprovalValidatorError(Exception):
    pass


class ApprovePrescription:
    prescription_model = "edc_pharmacy.rx"

    def __init__(self, rx_model_obj=None):
        dispense_appointment = rx_model_obj.dispense_appointment
        previous_appointment = dispense_appointment.previous()
        if (
            previous_appointment
            and rx_model_obj.__class__.objects.filter(
                is_approved=False, dispense_appointment=previous_appointment
            ).exists()
        ):
            raise PrescriptionApprovalValidatorError(
                f"Future prescriptions cannot be approved. Approve "
                f"prescriptions for dispensing date "
                f"{previous_appointment.appt_datetime}"
            )



# Generated by Django 4.1.2 on 2023-02-28 00:40

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0026_auto_20220826_0258"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataquery",
            name="dm_user",
            field=models.ForeignKey(
                blank=True,
                help_text="select a name from the list",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_data_manager.datamanageruser",
                verbose_name="DM resolved by",
            ),
        ),
    ]



# Generated by Django 2.1 on 2018-09-21 18:42

import _socket
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models

import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="DataRequest",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=25)),
                ("description", models.TextField(null=True)),
                ("decrypt", models.BooleanField(default=False)),
                (
                    "export_format",
                    models.CharField(choices=[("CSV", "CSV")], default="CSV", max_length=25),
                ),
                (
                    "models",
                    models.TextField(
                        help_text='List one table per line, no commas. Use "label lower" format.'
                    ),
                ),
            ],
            options={"ordering": ("name",)},
        ),
        migrations.CreateModel(
            name="DataRequestHistory",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("archive_filename", models.CharField(max_length=200, null=True)),
                ("emailed_to", models.EmailField(max_length=254, null=True)),
                ("emailed_datetime", models.DateTimeField(null=True)),
                ("summary", models.TextField(null=True)),
                (
                    "exported_datetime",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "data_request",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_pdutils.DataRequest",
                    ),
                ),
            ],
            options={"ordering": ("-exported_datetime",)},
        ),
        migrations.CreateModel(
            name="HistoricalDataRequest",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("name", models.CharField(max_length=25)),
                ("description", models.TextField(null=True)),
                ("decrypt", models.BooleanField(default=False)),
                (
                    "export_format",
                    models.CharField(choices=[("CSV", "CSV")], default="CSV", max_length=25),
                ),
                (
                    "models",
                    models.TextField(
                        help_text='List one table per line, no commas. Use "label lower" format.'
                    ),
                ),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        primary_key=True, serialize=False
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical data request",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



from django.contrib import admin
from edc_model.admin import HistoricalModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin

from ..admin_site import edc_qareports_admin
from ..models import Note
from .model_admin_mixin import ModelAdminMixin


@admin.register(Note.history.model, site=edc_qareports_admin)
class HistoricalNoteModelAdmin(
    HistoricalModelAdminMixin,
    ModelAdminMixin,
    SimpleHistoryAdmin,
):
    """A modeladmin class for the Note model."""

    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20
    change_list_note_url_name = "edc_qareports_admin:edc_qareports_note_changelist"

    list_display = (
        "subject_identifier",
        "report_model",
        "status",
        "history_type",
        "formatted_history_date",
        "revision",
        "history_id",
    )

    search_fields = ("report_model", "subject_identifier")



# Generated by Django 5.2.10 on 2026-01-24 15:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0107_remove_container_qty_remove_historicalcontainer_qty_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="container",
            name="qty_suggested",
            field=models.DecimalField(
                decimal_places=2,
                help_text="May be adjusted at receiving",
                max_digits=10,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="qty_suggested",
            field=models.DecimalField(
                decimal_places=2,
                help_text="May be adjusted at receiving",
                max_digits=10,
                null=True,
            ),
        ),
    ]



from .inline_hide_original_object_name_mixin import InlineHideOriginalObjectNameMixin
from .limited_admin_inline_mixin import LimitedAdminInlineMixin
from .stacked_inline_modeladmin_mixin import StackedInlineModelAdminMixin
from .tabular_inline_mixin import TabularInlineMixin



# Generated by Django 4.2.7 on 2023-12-05 03:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0030_remove_aliquot_edc_lab_ali_modifie_49b6da_idx_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="box",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Box",
                "verbose_name_plural": "Boxes",
            },
        ),
        migrations.AlterModelOptions(
            name="boxitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Box Item",
            },
        ),
        migrations.AlterModelOptions(
            name="boxtype",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Box Type",
            },
        ),
        migrations.AlterModelOptions(
            name="manifest",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Manifest",
            },
        ),
        migrations.AlterModelOptions(
            name="manifestitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Manifest Item",
            },
        ),
        migrations.AlterModelOptions(
            name="shipper",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Shipper",
            },
        ),
        migrations.AlterUniqueTogether(
            name="boxitem",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="manifestitem",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="panel",
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name="boxitem",
            constraint=models.UniqueConstraint(
                fields=("box", "position"), name="edc_lab_boxitem_box_pos_uniq"
            ),
        ),
        migrations.AddConstraint(
            model_name="boxitem",
            constraint=models.UniqueConstraint(
                fields=("box", "identifier"), name="edc_lab_boxitem_box_ide_uniq"
            ),
        ),
        migrations.AddConstraint(
            model_name="manifestitem",
            constraint=models.UniqueConstraint(
                fields=("manifest", "identifier"),
                name="edc_lab_manifestitem_manifest_uniq",
            ),
        ),
        migrations.AddConstraint(
            model_name="panel",
            constraint=models.UniqueConstraint(
                fields=("name", "lab_profile_name"), name="edc_lab_panel_name_uniq"
            ),
        ),
    ]
